# Copyright 2004-2011 Joe Wreschnig, Michael Urman, Iñigo Serna, Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import locale
import re

from quodlibet import _

from .iso639 import ISO_639_2


class ValidationError(Exception):
    pass


class Massager:
    """Massage a tag value from various 'okay' formats to the
    'correct' format."""

    tags: list[str] = []
    error = "Metaerror. This should be overridden in subclasses."
    options: list[str] = []

    _massagers: "dict[str, Massager]" = {}

    def validate(self, value):
        """Returns a validated value.

        Raises ValidationError if invalid.
        """

        raise NotImplementedError

    def is_valid(self, value):
        """Returns True if a field is valid, False if not"""

        try:
            self.validate(value)
        except ValidationError:
            return False
        return True

    @classmethod
    def _register(cls, other):
        """Register a new massager implementation"""

        assert issubclass(other, Massager)
        assert other.tags
        for tag in other.tags:
            cls._massagers[tag] = other
        return other

    @classmethod
    def get_massagers(cls):
        """Returns all massager subclasses"""

        return set(cls._massagers.values())

    @classmethod
    def for_tag(cls, tag):
        """Returns a massager instance for the tag or raises KeyError"""

        return cls._massagers[tag]()


def validate(tag, value):
    """Validate a value based on the tag.

    Raises ValidationError if invalid
    """

    try:
        return Massager.for_tag(tag).validate(value)
    except KeyError:
        return value


def is_valid(tag, value):
    """Returns True if the fields is valid"""

    try:
        return Massager.for_tag(tag).is_valid(value)
    except KeyError:
        return True


def error_message(tag, value):
    """Returns an error message for invalid tag values"""

    try:
        return Massager.for_tag(tag).error
    except KeyError:
        return ""


def get_options(tag):
    """Returns a list of suggested values for the tag. If the list is empty
    this either means that the tag is unknown or the set of valid values would
    be too large"""

    try:
        return list(Massager.for_tag(tag).options)
    except KeyError:
        return []


@Massager._register
class DateMassager(Massager):
    tags = ["date"]
    error = _(
        "The date must be entered in 'YYYY', 'YYYY-MM-DD' or "
        "'YYYY-MM-DD HH:MM:SS' format."
    )
    __match = re.compile(
        r"^\d{4}([-.]\d{2}([-.]\d{2}([T ]\d{2}([:.]\d{2}([:.]\d{2})?)?)?)?)?$"
    ).match

    def validate(self, value):
        value = value.strip().replace(".", "-").replace("/", "-")
        if not self.__match(value):
            raise ValidationError
        return value


@Massager._register
class GainMassager(Massager):
    tags = ["replaygain_album_gain", "replaygain_track_gain"]
    error = _("Replay Gain gains must be entered in 'x.yy dB' format.")
    __match = re.compile(r"^[+-]\d+\.?\d+?\s+dB$").match

    def validate(self, value):
        if self.__match(value):
            return value
        try:
            f = float(value.split()[0])
        except (IndexError, TypeError, ValueError):
            try:
                f = locale.atof(value.split()[0])
            except (IndexError, TypeError, ValueError) as e:
                raise ValidationError from e
        else:
            return (f"{f:+f}").rstrip("0") + " dB"


@Massager._register
class PeakMassager(Massager):
    tags = ["replaygain_album_peak", "replaygain_track_peak"]
    error = _("Replay Gain peaks must be entered in 'x.yy' format.")

    def validate(self, value):
        value = value.strip()
        try:
            f = float(value)
        except (TypeError, ValueError):
            try:
                f = locale.atof(value)
            except (TypeError, ValueError) as e:
                raise ValidationError from e
        else:
            if f < 0 or f >= 2:
                raise ValidationError
            return str(f)


@Massager._register
class MBIDMassager(Massager):
    tags = [
        "musicbrainz_trackid",
        "musicbrainz_albumid",
        "musicbrainz_artistid",
        "musicbrainz_albumartistid",
        "musicbrainz_trmid",
        "musicip_puid",
    ]
    error = _("MusicBrainz IDs must be in UUID format.")

    def validate(self, value):
        value = value.encode("ascii", "replace").decode("ascii")
        value = "".join(filter(str.isalnum, value.strip().lower()))
        try:
            int(value, 16)
        except ValueError as e:
            raise ValidationError from e
        else:
            if len(value) != 32:
                raise ValidationError
            return "-".join(
                [value[:8], value[8:12], value[12:16], value[16:20], value[20:]]
            )


@Massager._register
class MBAlbumStatus(Massager):
    tags = ["musicbrainz_albumstatus"]
    # Translators: Leave "official", "promotional", and "bootleg"
    # untranslated. They are the three possible literal values.
    error = _(
        "MusicBrainz release status must be 'official', 'promotional', or 'bootleg'."
    )
    options = ["official", "promotional", "bootleg"]

    def validate(self, value):
        if value not in self.options:
            raise ValidationError
        return value


@Massager._register
class LanguageMassager(Massager):
    tags = ["language"]
    error = _("Language must be an ISO 639-2 three-letter code")

    options = ISO_639_2

    def validate(self, value):
        # Issue 439: Actually, allow free-text through
        return value

    def is_valid(self, value):
        # Override, to allow empty string to be a valid language (freetext)
        return True



# Copyright 2004-2012 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2011-2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
from collections.abc import Sequence

from gi.repository import Gtk, Pango, Gdk

from quodlibet import C_, _, ngettext, print_e, print_d
from quodlibet import app
from quodlibet import config
from quodlibet import qltk
from quodlibet import util
from quodlibet.formats import AudioFileError
from quodlibet.plugins import PluginManager
from quodlibet.plugins.editing import EditTagsPlugin
from quodlibet.qltk import Icons
from quodlibet.qltk._editutils import EditingPluginHandler, OverwriteWarning
from quodlibet.qltk._editutils import WriteFailedError
from quodlibet.qltk.ccb import ConfigCheckButton
from quodlibet.qltk.completion import LibraryValueCompletion
from quodlibet.qltk.models import ObjectStore
from quodlibet.qltk.tagscombobox import TagsComboBox, TagsComboBoxEntry
from quodlibet.qltk.views import RCMHintedTreeView, TreeViewColumn, BaseView
from quodlibet.qltk.window import Dialog
from quodlibet.qltk.wlw import WritingWindow
from quodlibet.qltk.x import SeparatorMenuItem, Button, MenuItem
from quodlibet.util import connect_obj
from quodlibet.util import massagers
from quodlibet.util.i18n import numeric_phrase
from quodlibet.util.string.splitters import (
    split_value,
    split_title,
    split_people,
    split_album,
)
from quodlibet.util.tags import USER_TAGS, MACHINE_TAGS, sortkey as tagsortkey


class Comment:
    """A summary of a collection of values for one tag"""

    def __init__(self, text, have=1, total=1, shared=True):
        """
        Args:
            text: the first or only text value
            have: amount of songs that have a value
            total: total amount of songs
            shared: if all songs that have a value, have the same one
        """

        self.complete = have == total
        self.shared = shared
        self.total = total
        self.missing = total - have
        self.have = have
        self.text = text

    def _paren(self):
        if self.shared:
            return numeric_phrase(
                "missing from %d song", "missing from %d songs", self.missing
            )
        if self.complete:
            return numeric_phrase(
                "different across %d song", "different across %d songs", self.total
            )
        d = numeric_phrase(
            "different across %d song", "different across %d songs", self.have
        )
        m = numeric_phrase(
            "missing from %d song", "missing from %d songs", self.missing
        )
        return f"{d}, {m}"

    def is_special(self):
        return not self.shared or not self.complete

    def is_missing(self):
        return not self.complete

    def get_shared_text(self):
        if self.shared:
            return util.escape(self.text)
        return ""

    def get_markup(self):
        """Returns pango markup for displaying"""

        if self.shared and self.complete:
            return util.escape(self.text)
        if self.shared:
            return "\n".join(
                [
                    f"{util.escape(s)}{util.italic(' ' + self._paren())}"
                    for s in self.text.split("\n")
                ]
            )
        return util.italic(self._paren())


def get_default_tags():
    """Returns a list of tags that should be displayed even if not present
    in the file.
    """

    text = config.get("editing", "default_tags").strip()
    if not text:
        return []
    return text.split(",")


class AudioFileGroup(dict):
    """Values are a list of Comment instances"""

    def __init__(self, songs, real_keys_only=True):
        keys = {}
        first = {}
        all = {}
        total = len(songs)
        self.songs = songs
        self.is_file = True
        can_multi = True
        can_change = True

        for song in songs:
            self.is_file &= song.is_file

            if real_keys_only:
                iter_func = song.iterrealitems
            else:
                iter_func = song.items

            for comment, val in iter_func():
                keys[comment] = keys.get(comment, 0) + 1
                first.setdefault(comment, val)
                all[comment] = all.get(comment, True) and first[comment] == val

            song_can_multi = song.can_multiple_values()
            if song_can_multi is not True:
                if can_multi is True:
                    can_multi = set(song_can_multi)
                else:
                    can_multi.intersection_update(song_can_multi)

            song_can_change = song.can_change()
            if song_can_change is not True:
                if can_change is True:
                    can_change = set(song_can_change)
                else:
                    can_change.intersection_update(song_can_change)

        self._can_multi = can_multi
        self._can_change = can_change

        # collect comment representations
        for tag, count in keys.items():
            first_value = first[tag]
            if not isinstance(first_value, str):
                first_value = str(first_value)
            shared = all[tag]
            complete = count == total
            if shared and complete:
                values = first_value.split("\n")
            else:
                values = [first_value]
            for v in values:
                self.setdefault(tag, []).append(Comment(v, count, total, shared))

    def can_multiple_values(self, key=None):
        """If no arguments passed returns a set of tags that have multi
        value support for all contained songs. If key is given returns
        if all songs support multi value tags for that key.
        """

        if key is None:
            return self._can_multi
        return all(song.can_multiple_values(key) for song in self.songs)

    def can_change(self, key=None):
        """See can_multiple_values()"""

        if key is None:
            return self._can_change
        return all(song.can_change(key) for song in self.songs)


class TagSplitter(EditTagsPlugin):
    """Splits tag values into other tags"""


class SplitValues(TagSplitter):
    def __init__(self, tag, value):
        super().__init__(label=_("Split into _Multiple Values"), use_underline=True)
        self.set_image(
            Gtk.Image.new_from_icon_name(Icons.EDIT_FIND_REPLACE, Gtk.IconSize.MENU)
        )
        spls = config.gettext("editing", "split_on").split()
        vals = [
            val if len(val) <= 64 else val[:64] + "…"
            for val in split_value(value, spls)
        ]
        string = ", ".join([f"{tag}={val}" for val in vals])
        self.set_label(string)
        self.set_sensitive(len(vals) > 1)

    def activated(self, tag, value):
        spls = config.gettext("editing", "split_on").split()
        return [(tag, v) for v in split_value(value, spls)]


class SplitDisc(TagSplitter):
    tags = ["album"]
    needs = ["discnumber"]
    _order = 0.5

    def __init__(self, tag, value):
        super().__init__(label=_("Split Disc out of _Album"), use_underline=True)
        self.set_image(
            Gtk.Image.new_from_icon_name(Icons.EDIT_FIND_REPLACE, Gtk.IconSize.MENU)
        )

        album, disc = split_album(value)
        if disc is not None:
            album = album if len(album) <= 64 else album[:64] + "…"
            self.set_label(f"{tag}={album}, {self.needs[0]}={disc}")

        self.set_sensitive(disc is not None)

    def activated(self, tag, value):
        album, disc = split_album(value)
        return [(tag, album), ("discnumber", disc)]


class SplitTitle(TagSplitter):
    tags = ["title"]
    needs = ["version"]
    _order = 0.5

    def __init__(self, tag, value):
        super().__init__(label=_("Split _Version out of Title"), use_underline=True)
        self.set_image(
            Gtk.Image.new_from_icon_name(Icons.EDIT_FIND_REPLACE, Gtk.IconSize.MENU)
        )
        tag_spls = config.gettext("editing", "split_on").split()
        sub_spls = config.gettext("editing", "sub_split_on").split()

        title, versions = split_title(value, tag_spls, sub_spls)
        if versions:
            title = title if len(title) <= 64 else title[:64] + "…"
            versions = [ver if len(ver) <= 64 else ver[:64] + "…" for ver in versions]
            string = ", ".join(
                [f"{tag}={title}"] + [f"{self.needs[0]}={ver}" for ver in versions]
            )
            self.set_label(string)

        self.set_sensitive(bool(versions))

    def activated(self, tag, value):
        tag_spls = config.gettext("editing", "split_on").split()
        sub_spls = config.gettext("editing", "sub_split_on").split()
        title, versions = split_title(value, tag_spls, sub_spls)
        return [(tag, title)] + [("version", v) for v in versions]


class SplitPerson(TagSplitter):
    tags = ["artist"]
    _order = 0.5

    def __init__(self, tag, value):
        super().__init__(label=self.title, use_underline=True)
        self.set_image(
            Gtk.Image.new_from_icon_name(Icons.EDIT_FIND_REPLACE, Gtk.IconSize.MENU)
        )
        tag_spls = config.gettext("editing", "split_on").split()
        sub_spls = config.gettext("editing", "sub_split_on").split()

        artist, others = split_people(value, tag_spls, sub_spls)
        if others:
            artist = artist if len(artist) <= 64 else artist[:64] + "…"
            others = [
                other if len(other) <= 64 else other[:64] + "…" for other in others
            ]
            string = ", ".join(
                [f"{tag}={artist}"] + [f"{self.needs[0]}={o}" for o in others]
            )
            self.set_label(string)

        self.set_sensitive(bool(others))

    def activated(self, tag, value):
        tag_spls = config.gettext("editing", "split_on").split()
        sub_spls = config.gettext("editing", "sub_split_on").split()
        artist, others = split_people(value, tag_spls, sub_spls)
        return [(tag, artist)] + [(self.needs[0], o) for o in others]


class SplitArranger(SplitPerson):
    needs = ["arranger"]
    title = _("Split Arranger out of Ar_tist")


class SplitPerformer(SplitPerson):
    needs = ["performer"]
    title = _("Split _Performer out of Artist")


class SplitPerformerFromTitle(SplitPerson):
    tags = ["title"]
    needs = ["performer"]
    title = _("Split _Performer out of Title")


class SplitOriginalArtistFromTitle(SplitPerson):
    tags = ["title"]
    needs = ["originalartist"]
    title = _("Split _Originalartist out of Title")


class AddTagDialog(Dialog):
    def __init__(self, parent, can_change, library):
        super().__init__(
            title=_("Add a Tag"),
            transient_for=qltk.get_top_parent(parent),
            use_header_bar=True,
        )
        self.set_border_width(6)
        self.set_resizable(False)
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        add = self.add_icon_button(_("_Add"), Icons.LIST_ADD, Gtk.ResponseType.OK)
        self.vbox.set_spacing(6)
        self.set_default_response(Gtk.ResponseType.OK)
        table = Gtk.Table(n_rows=2, n_columns=2)
        table.set_row_spacings(12)
        table.set_col_spacings(6)
        table.set_border_width(6)

        self.__tag = (
            TagsComboBoxEntry() if can_change is True else TagsComboBox(can_change)
        )

        label = Gtk.Label()
        label.set_alignment(0.0, 0.5)
        label.set_text(_("_Tag:"))
        label.set_use_underline(True)
        label.set_mnemonic_widget(self.__tag)
        table.attach(label, 0, 1, 0, 1)
        table.attach(self.__tag, 1, 2, 0, 1)

        self.__val = Gtk.Entry()
        self.__val.set_completion(LibraryValueCompletion("", library))
        label = Gtk.Label()
        label.set_text(_("_Value:"))
        label.set_alignment(0.0, 0.5)
        label.set_use_underline(True)
        label.set_mnemonic_widget(self.__val)
        valuebox = Gtk.EventBox()
        table.attach(label, 0, 1, 1, 2)
        table.attach(valuebox, 1, 2, 1, 2)
        hbox = Gtk.HBox()
        valuebox.add(hbox)
        hbox.pack_start(self.__val, True, True, 0)
        hbox.set_spacing(6)
        invalid = Gtk.Image.new_from_icon_name(
            Icons.DIALOG_WARNING, Gtk.IconSize.SMALL_TOOLBAR
        )
        hbox.pack_start(invalid, True, True, 0)

        self.vbox.pack_start(table, True, True, 0)
        self.get_child().show_all()
        invalid.hide()

        for entry in [self.__tag, self.__val]:
            entry.connect("changed", self.__validate, add, invalid, valuebox)
        self.__tag.connect("changed", self.__set_value_completion, library)
        self.__set_value_completion(self.__tag, library)

        if can_change is True:
            connect_obj(
                self.__tag.get_child(), "activate", Gtk.Entry.grab_focus, self.__val
            )

    def __set_value_completion(self, tag, library):
        completion = self.__val.get_completion()
        if completion:
            completion.set_tag(self.__tag.tag, library)

    def get_tag(self):
        try:
            return self.__tag.tag
        except AttributeError:
            return self.__tag.tag

    def get_value(self):
        return self.__val.get_text()

    def __validate(self, editable, add, invalid, box):
        tag = self.get_tag()
        value = self.get_value()
        valid = massagers.is_valid(tag, value)
        add.set_sensitive(valid)
        if valid:
            invalid.hide()
            box.set_tooltip_text("")
        else:
            invalid.show()
            box.set_tooltip_text(massagers.error_message(tag, value))

    def run(self):
        self.show()
        self.__val.set_activates_default(True)
        self.__tag.grab_focus()
        return super().run()


class EditTagsPluginHandler(EditingPluginHandler):
    from quodlibet.plugins.editing import EditTagsPlugin

    Kind = EditTagsPlugin


class ListEntry:
    """Holds a Comment and some state for the editing process"""

    tag = None
    value = None
    edited = False
    canedit = True
    deleted = False
    origvalue = None
    renamed = False
    origtag = None

    def __init__(self, tag, value):
        self.tag = tag
        self.value = value


class EditTags(Gtk.VBox):
    handler = EditTagsPluginHandler()

    _SPLITTERS: Sequence[type[TagSplitter]] = sorted(
        [
            SplitDisc,
            SplitTitle,
            SplitPerformer,
            SplitArranger,
            SplitValues,
            SplitPerformerFromTitle,
            SplitOriginalArtistFromTitle,
        ],
        key=lambda item: (item._order, item.__name__),
    )

    @classmethod
    def init_plugins(cls):
        PluginManager.instance.register_handler(cls.handler)

    def __init__(self, parent, library):
        super().__init__(spacing=12)
        self.title = _("Edit Tags")
        self.set_border_width(12)
        self._group_info = None

        model = ObjectStore()
        view = RCMHintedTreeView(model=model)
        self._view = view
        selection = view.get_selection()
        render = Gtk.CellRendererPixbuf()
        column = TreeViewColumn()
        column.pack_start(render, True)
        column.set_fixed_width(24)
        column.set_expand(False)

        def cdf_write(col, rend, model, iter_, *args):
            entry = model.get_value(iter_)
            rend.set_property("sensitive", entry.edited or entry.deleted)
            if entry.canedit or entry.deleted:
                if entry.deleted:
                    rend.set_property("icon-name", Icons.EDIT_DELETE)
                else:
                    rend.set_property("icon-name", Icons.EDIT)
            else:
                rend.set_property("icon-name", Icons.CHANGES_PREVENT)

        column.set_cell_data_func(render, cdf_write)
        view.append_column(column)

        render = Gtk.CellRendererText()
        column = TreeViewColumn(title=_("Tag"))
        column.pack_start(render, True)

        def cell_data_tag(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            cell.set_property("text", entry.tag)
            cell.set_property("strikethrough", entry.deleted)

        column.set_cell_data_func(render, cell_data_tag)

        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        render.set_property("editable", True)
        render.connect("edited", self.__edit_tag_name, model)
        render.connect("editing-started", self.__tag_editing_started, model, library)
        view.append_column(column)

        render = Gtk.CellRendererText()
        render.set_property("ellipsize", Pango.EllipsizeMode.END)
        render.set_property("editable", True)
        render.connect("edited", self.__edit_tag, model)
        render.connect("editing-started", self.__value_editing_started, model, library)
        column = TreeViewColumn(title=_("Value"))
        column.pack_start(render, True)

        def cell_data_value(column, cell, model, iter_, data):
            entry = model.get_value(iter_)
            markup = entry.value.get_markup()
            cell.markup = markup
            cell.set_property("markup", markup)
            cell.set_property("editable", entry.canedit)
            cell.set_property("strikethrough", entry.deleted)

        column.set_cell_data_func(render, cell_data_value)

        column.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        view.append_column(column)

        sw = Gtk.ScrolledWindow()
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        sw.add(view)
        self.pack_start(sw, True, True, 0)

        cb = ConfigCheckButton(
            _("Show _programmatic tags"),
            "editing",
            "alltags",
            populate=True,
            tooltip=_(
                "Access all tags, including machine-generated "
                "ones e.g. MusicBrainz or Replay Gain tags"
            ),
        )
        cb.connect("toggled", self.__checkbox_toggled)
        hb = Gtk.HBox()
        hb.pack_start(cb, False, True, 0)

        cb = ConfigCheckButton(
            _("Show _multi-line tags"),
            "editing",
            "show_multi_line_tags",
            populate=True,
            tooltip=_("Show potentially multi-line tags (e.g 'lyrics') here too"),
        )
        cb.connect("toggled", self.__checkbox_toggled)
        hb.pack_start(cb, False, True, 12)
        self.pack_start(hb, False, True, 0)

        # Add and Remove [tags] buttons
        buttonbox = Gtk.HBox(spacing=18)
        bbox1 = Gtk.HButtonBox()
        bbox1.set_spacing(6)
        bbox1.set_layout(Gtk.ButtonBoxStyle.START)
        add = qltk.Button(_("_Add…"), Icons.LIST_ADD)
        add.set_focus_on_click(False)
        self._add = add
        add.connect("clicked", self.__add_tag, model, library)
        bbox1.pack_start(add, True, True, 0)
        # Remove button
        remove = qltk.Button(_("_Remove"), Icons.LIST_REMOVE)
        remove.set_focus_on_click(False)
        remove.connect("clicked", self.__remove_tag, view)
        remove.set_sensitive(False)
        self._remove = remove

        bbox1.pack_start(remove, True, True, 0)

        # Revert and save buttons
        # Both can have customised translated text (and thus accels)
        bbox2 = Gtk.HButtonBox()
        bbox2.set_spacing(6)
        bbox2.set_layout(Gtk.ButtonBoxStyle.END)
        # Translators: Revert button in the tag editor
        revert = Button(C_("edittags", "_Revert"), Icons.DOCUMENT_REVERT)

        self._revert = revert
        revert.set_sensitive(False)
        # Translators: Save button in the tag editor
        save = Button(C_("edittags", "_Save"), Icons.DOCUMENT_SAVE)
        save.set_sensitive(False)
        self._save = save
        bbox2.pack_start(revert, True, True, 0)
        bbox2.pack_start(save, True, True, 0)

        buttonbox.pack_start(bbox1, True, True, 0)
        buttonbox.pack_start(bbox2, True, True, 0)
        self.pack_start(buttonbox, False, True, 0)
        self._buttonbox = buttonbox

        parent.connect("changed", self.__parent_changed)
        revert.connect("clicked", lambda *x: self._update())
        connect_obj(revert, "clicked", parent.set_pending, None)

        save.connect("clicked", self.__save_files, revert, model, library)
        connect_obj(save, "clicked", parent.set_pending, None)
        for sig in ["row-inserted", "row-deleted", "row-changed"]:
            model.connect(sig, self.__enable_save, [save, revert])
            connect_obj(model, sig, parent.set_pending, save)

        view.connect("popup-menu", self._popup_menu, parent)
        view.connect("button-press-event", self.__button_press)
        view.connect("key-press-event", self.__view_key_press_event)
        selection.connect("changed", self.__tag_select, remove)
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)

        self._parent = parent

        for child in self.get_children():
            child.show_all()

    def __checkbox_toggled(self, *args):
        self._update()

    def __view_key_press_event(self, view, event):
        if qltk.is_accel(event, "Delete"):
            self.__remove_tag(view, view)
            return Gdk.EVENT_STOP
        if qltk.is_accel(event, "<Primary>s"):
            # Issue 697: allow Ctrl-s to save.
            self._save.emit("clicked")
            return Gdk.EVENT_STOP
        if qltk.is_accel(event, "<Primary>c"):
            self.__copy_tag_value(event, view)
            return Gdk.EVENT_STOP
        return Gdk.EVENT_PROPAGATE

    def __enable_save(self, *args):
        buttons = args[-1]
        for b in buttons:
            b.set_sensitive(True)

    def __paste(self, clip, text, args):
        rend, path = args
        if text:
            rend.emit("edited", path, text.strip())

    def __menu_activate(self, activator, view):
        model, (path,) = view.get_selection().get_selected_rows()
        entry = model[path][0]

        tag = entry.tag
        comment = entry.value
        value = comment.text
        vals = activator.activated(tag, value)
        replaced = False
        if vals and (len(vals) != 1 or vals[0][1] != value):
            for atag, aval in vals:
                if atag == tag and not replaced:
                    replaced = True
                    entry.value = Comment(aval)
                    entry.edited = True
                    model.path_changed(path)
                else:
                    self.__add_new_tag(model, atag, aval)
        elif vals:
            replaced = True

        if not replaced:
            entry.edited = entry.deleted = True
            model.path_changed(path)

    def __item_for(
        self, view: BaseView, item_cls: type[EditTagsPlugin], tag: str, text: str
    ) -> EditTagsPlugin | None:
        try:
            item = item_cls(tag, text)
        except Exception as e:
            print_e(f"Couldn't create menu item from {item_cls} ({e})")
            return None
        else:
            item.connect("activate", self.__menu_activate, view)
            return item

    def _popup_menu(self, view: BaseView, _parent):
        menu = Gtk.Menu()

        view.ensure_popup_selection()
        model, rows = view.get_selection().get_selected_rows()
        can_change = all(model[path][0].canedit for path in rows)

        if len(rows) == 1:
            row = model[rows[0]]
            entry = row[0]

            comment = entry.value
            text = comment.text

            split_menu = Gtk.Menu()

            for Item in self._SPLITTERS:
                if Item.tags and entry.tag not in Item.tags:
                    continue
                item = self.__item_for(view, Item, entry.tag, text)
                if not item:
                    continue
                vals = item.activated(entry.tag, text)
                changeable = any(not self._group_info.can_change(k) for k in item.needs)
                fixed = changeable or comment.is_special()
                if fixed:
                    item.set_sensitive(False)
                if len(vals) > 1 and vals[1][1]:
                    split_menu.append(item)
            if split_menu.get_children():
                split_menu.append(SeparatorMenuItem())

            plugins = self.handler.plugins
            print_d(
                f"Adding {len(plugins)} plugin(s) to menu: "
                f"{', '.join(p.__name__ for p in plugins)}"
            )
            for p_cls in plugins:
                item = self.__item_for(view, p_cls, entry.tag, text)
                if not item:
                    continue
                results = item.activated(entry.tag, text)
                # Only enable for the user if the plugin would do something
                item.set_sensitive(results != [(entry.tag, text)])
                menu.append(item)
            pref_item = MenuItem(_("_Configure"), Icons.PREFERENCES_SYSTEM)
            split_menu.append(pref_item)

            def show_prefs(parent):
                from quodlibet.qltk.exfalsowindow import ExFalsoWindow

                if isinstance(app.window, ExFalsoWindow):
                    from quodlibet.qltk.exfalsowindow import PreferencesWindow

                    window = PreferencesWindow(parent)
                else:
                    from quodlibet.qltk.prefs import PreferencesWindow

                    window = PreferencesWindow(parent, open_page="tagging")
                window.show()

            connect_obj(pref_item, "activate", show_prefs, self)

            split_item = MenuItem(_("_Split Tag"), Icons.EDIT_FIND_REPLACE)

            if split_menu.get_children():
                split_item.set_submenu(split_menu)
            else:
                split_item.set_sensitive(False)

            menu.append(split_item)

        copy_b = MenuItem(_("_Copy Value(s)"), Icons.EDIT_COPY)
        copy_b.connect("activate", self.__copy_tag_value, view)
        qltk.add_fake_accel(copy_b, "<Primary>c")
        menu.append(copy_b)

        remove_b = MenuItem(_("_Remove"), Icons.LIST_REMOVE)
        remove_b.connect("activate", self.__remove_tag, view)
        qltk.add_fake_accel(remove_b, "Delete")
        menu.append(remove_b)

        menu.show_all()
        # Setting the menu itself to be insensitive causes it to not
        # be dismissed; see #473.
        for c in menu.get_children():
            c.set_sensitive(can_change and c.get_property("sensitive"))
        copy_b.set_sensitive(True)
        remove_b.set_sensitive(True)
        menu.connect("selection-done", lambda m: m.destroy())

        # XXX: Keep reference
        self.__menu = menu
        return view.popup_menu(menu, 3, Gtk.get_current_event_time())

    def __tag_select(self, selection, remove):
        model, rows = selection.get_selected_rows()
        remove.set_sensitive(bool(rows))

    def __add_new_tag(self, model, tag, value):
        assert isinstance(value, str)
        iters = [i for (i, v) in model.iterrows() if v.tag == tag]
        if iters and not self._group_info.can_multiple_values(tag):
            title = _("Unable to add tag")
            msg = _("Unable to add %s") % util.bold(tag)
            msg += "\n\n"
            msg += _(
                "The files currently selected do not support multiple values for %s."
            ) % util.bold(tag)
            qltk.ErrorMessage(self, title, msg, escape_desc=False).run()
            return

        entry = ListEntry(tag, Comment(value))
        entry.edited = True

        if len(iters):
            model.insert_after(iters[-1], row=[entry])
        else:
            model.append(row=[entry])

    def __add_tag(self, activator, model, library):
        add = AddTagDialog(self, self._group_info.can_change(), library)

        while True:
            resp = add.run()
            if resp != Gtk.ResponseType.OK:
                break
            tag = add.get_tag()
            value = add.get_value()
            assert isinstance(value, str)
            value = massagers.validate(tag, value)
            assert isinstance(value, str)
            if not self._group_info.can_change(tag):
                title = ngettext("Invalid tag", "Invalid tags", 1)
                msg = ngettext(
                    "Invalid tag %s\n\nThe files currently "
                    "selected do not support editing this tag.",
                    "Invalid tags %s\n\nThe files currently "
                    "selected do not support editing these tags.",
                    1,
                ) % util.bold(tag)
                qltk.ErrorMessage(self, title, msg).run()
            else:
                self.__add_new_tag(model, tag, value)
                break

        add.destroy()

    def __remove_tag(self, activator, view):
        model, paths = view.get_selection().get_selected_rows()
        # Since the iteration can modify path numbers, we need accurate
        # rows (= iters) before we start.
        rows = [model[path] for path in paths]
        for row in rows:
            entry = row[0]
            if entry.origvalue is not None:
                entry.edited = entry.deleted = True
                model.row_changed(row.path, row.iter)
            else:
                model.remove(row.iter)

    def __copy_tag_value(self, activator, view):
        model, paths = view.get_selection().get_selected_rows()
        rows = [model[path] for path in paths]
        values = []
        for row in rows:
            entry_text = row[0].value.get_shared_text()
            if entry_text:
                values.append(entry_text)
        text = "\n".join(values)
        if len(text) > 0:
            clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
            clipboard.set_text(text, -1)

    def __save_files(self, save, revert, model, library):
        updated = {}
        deleted = {}
        added = {}
        renamed = {}

        for entry in model.values():
            if entry.edited and not (entry.deleted or entry.renamed):
                if entry.origvalue is not None:
                    l = updated.setdefault(entry.tag, [])
                    l.append((entry.value, entry.origvalue))
                else:
                    l = added.setdefault(entry.tag, [])
                    l.append(entry.value)

            if entry.edited and entry.deleted:
                if entry.origvalue is not None:
                    l = deleted.setdefault(entry.tag, [])
                    l.append(entry.origvalue)

            if entry.edited and entry.renamed and not entry.deleted:
                l = renamed.setdefault(entry.tag, [])
                l.append((entry.origtag, entry.value, entry.origvalue))

        was_changed = set()
        songs = self._group_info.songs
        win = WritingWindow(self, len(songs))
        win.show()
        all_done = False
        for song in songs:
            if not song.valid():
                win.hide()
                dialog = OverwriteWarning(self, song)
                resp = dialog.run()
                win.show()
                if resp != OverwriteWarning.RESPONSE_SAVE:
                    break

            changed = False
            for key, values in updated.items():
                for new_value, old_value in values:
                    if song.can_change(key):
                        if old_value is None:
                            song.add(key, new_value.text)
                        else:
                            song.change(key, old_value.text, new_value.text)
                        changed = True

            for key, values in added.items():
                for value in values:
                    if song.can_change(key):
                        song.add(key, value.text)
                        changed = True

            for key, values in deleted.items():
                for value in values:
                    if not value.shared:
                        # In case it isn't shared we don't know the actual
                        # values to remove. But we know that in that case
                        # we merge all values into one Comment so just removing
                        # everything for that key is OK.
                        song.remove(key, None)
                        changed = True
                    elif key in song:
                        song.remove(key, value.text)
                        changed = True

            save_rename = []
            for new_tag, values in renamed.items():
                for old_tag, new_value, old_value in values:
                    if song.can_change(new_tag) and old_tag in song:
                        if not new_value.is_special():
                            song.remove(old_tag, old_value.text)
                            save_rename.append((new_tag, new_value))
                        elif new_value.is_missing():
                            song.remove(old_tag, old_value)
                            save_rename.append((new_tag, new_value))
                        else:
                            save_rename.append((new_tag, Comment(song[old_tag])))
                            song.remove(old_tag, None)
                        changed = True

            for tag, value in save_rename:
                song.add(tag, value.text)

            if changed:
                try:
                    song.write()
                except AudioFileError:
                    util.print_exc()
                    WriteFailedError(self, song).run()
                    library.reload(song, changed=was_changed)
                    break
                was_changed.add(song)

            if win.step():
                break
        else:
            all_done = True

        win.destroy()
        library.changed(was_changed)
        for b in [save, revert]:
            b.set_sensitive(not all_done)

    def __edit_tag(self, renderer, path, new_value, model):
        path = Gtk.TreePath.new_from_string(path)
        entry = model[path][0]
        error_dialog = None

        if not massagers.is_valid(entry.tag, new_value):
            error_dialog = qltk.WarningMessage(
                self,
                _("Invalid value"),
                _("Invalid value: %(value)s\n\n%(error)s")
                % {
                    "value": util.bold(new_value),
                    "error": massagers.error_message(entry.tag, new_value),
                },
            )
        else:
            new_value = massagers.validate(entry.tag, new_value)

        comment = entry.value
        changed = comment.text != new_value
        identical = comment.shared and comment.complete
        if (changed and (identical or new_value)) or (new_value and not identical):
            # only give an error if we would have applied the value
            if error_dialog is not None:
                error_dialog.run()
                return
            entry.value = Comment(new_value)
            entry.edited = True
            entry.deleted = False
            model.path_changed(path)

    def __edit_tag_name(self, renderer, path, new_tag, model):
        new_tag = " ".join(new_tag.splitlines()).lower()
        path = Gtk.TreePath.new_from_string(path)
        entry = model[path][0]
        if new_tag == entry.tag:
            return
        if not self._group_info.can_change(new_tag):
            # Can't add the new tag.
            title = ngettext("Invalid tag", "Invalid tags", 1)
            msg = ngettext(
                "Invalid tag %s\n\nThe files currently "
                "selected do not support editing this tag.",
                "Invalid tags %s\n\nThe files currently "
                "selected do not support editing these tags.",
                1,
            ) % util.bold(new_tag)
            qltk.ErrorMessage(self, title, msg, escape_desc=False).run()
        else:
            # FIXME: In case this is a special one we only
            # validate one value and never write it back..

            text = entry.value.text
            if not massagers.is_valid(new_tag, text):
                qltk.WarningMessage(
                    self,
                    _("Invalid value"),
                    _("Invalid value: %(value)s\n\n%(error)s")
                    % {
                        "value": util.bold(text),
                        "error": massagers.error_message(new_tag, text),
                    },
                ).run()
                return
            text = massagers.validate(new_tag, text)

            if entry.origvalue is None:
                # The tag hasn't been saved yet, so we can just update
                # the name in the model, and the value, since it
                # may have been re-validated.
                entry.tag = new_tag
                entry.value = Comment(text)
            else:
                # The tag has been saved, so delete the old tag and
                # add a new one with the old (or sanitized) value.
                entry.renamed = entry.edited = True
                entry.origtag = entry.tag
                entry.tag = new_tag
                if not entry.value.is_special():
                    entry.value = Comment(text)

            entry.canedit = True

            model.row_changed(path, model.get_iter(path))

    def __button_press(self, view, event):
        if event.button not in [Gdk.BUTTON_PRIMARY, Gdk.BUTTON_MIDDLE]:
            return Gdk.EVENT_PROPAGATE

        x, y = map(int, [event.x, event.y])
        try:
            path, col, cellx, celly = view.get_path_at_pos(x, y)
        except TypeError:
            return Gdk.EVENT_PROPAGATE

        if event.button == Gdk.BUTTON_MIDDLE and col == view.get_columns()[2]:
            display = Gdk.DisplayManager.get().get_default_display()
            selection = Gdk.SELECTION_PRIMARY
            if sys.platform == "win32":
                selection = Gdk.SELECTION_CLIPBOARD

            clipboard = Gtk.Clipboard.get_for_display(display, selection)
            for rend in col.get_cells():
                if rend.get_property("editable"):
                    clipboard.request_text(self.__paste, (rend, path.get_indices()[0]))
                    return Gdk.EVENT_STOP
            else:
                return Gdk.EVENT_PROPAGATE
        else:
            return Gdk.EVENT_PROPAGATE

    def _update(self, songs=None):
        if songs is None:
            songs = self._group_info.songs
        else:
            self._group_info = AudioFileGroup(songs)
        info = self._group_info

        keys = list(info.keys())
        default_tags = get_default_tags()
        keys = set(keys + default_tags)

        def custom_sort(key):
            try:
                prio = default_tags.index(key)
            except ValueError:
                prio = len(default_tags)
            return (prio, tagsortkey(key))

        if not config.getboolean("editing", "alltags"):
            keys = filter(lambda k: k not in MACHINE_TAGS, keys)

        if not config.getboolean("editing", "show_multi_line_tags"):
            tags = config.getstringlist("editing", "multi_line_tags")
            keys = filter(lambda k: k not in tags, keys)

        if not songs:
            keys = []

        with self._view.without_model() as model:
            model.clear()

            for tag in sorted(keys, key=custom_sort):
                canedit = info.can_change(tag)

                # default tags
                if tag not in info:
                    entry = ListEntry(tag, Comment(""))
                    entry.canedit = canedit
                    model.append(row=[entry])
                    continue

                for value in info[tag]:
                    entry = ListEntry(tag, value)
                    entry.origvalue = value
                    entry.edited = False
                    entry.canedit = canedit
                    entry.deleted = False
                    entry.renamed = False
                    entry.origtag = ""
                    model.append(row=[entry])

        self._buttonbox.set_sensitive(bool(info.can_change()))
        self._revert.set_sensitive(False)
        self._remove.set_sensitive(False)
        self._save.set_sensitive(False)
        self._add.set_sensitive(bool(songs))
        self._parent.set_pending(None)

    def __parent_changed(self, parent, songs):
        self._update(songs)

    def __value_editing_started(self, render, editable, path, model, library):
        if not editable.get_completion():
            tag = model[path][0].tag
            completion = LibraryValueCompletion(tag, library)
            editable.set_completion(completion)

        if isinstance(editable, Gtk.Entry):
            comment = model[path][0].value
            if comment.shared:
                editable.set_text(comment.text)
            else:
                editable.set_text("")

    def __tag_editing_started(self, render, editable, path, model, library):
        if not editable.get_completion():
            tags = self._group_info.can_change()
            if tags is True:
                tags = USER_TAGS
            completion = qltk.EntryCompletion(tags)
            editable.set_completion(completion)



# Copyright 2004-2005 Joe Wreschnig, Niklas Janlert
#           2012 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys

from senf import fsnative

from quodlibet import _
from quodlibet import app
from quodlibet import util
from quodlibet import const
from quodlibet import config


def main(argv=None):
    if argv is None:
        argv = sys.argv

    import quodlibet

    config_file = os.path.join(quodlibet.get_user_dir(), "config")
    quodlibet.init(config_file=config_file)

    from quodlibet.qltk import add_signal_watch

    add_signal_watch(app.quit)

    opts = util.OptionParser(
        "Ex Falso",
        const.VERSION,
        _("an audio tag editor"),
        "[{}]".format(_("directory")),
    )

    argv.append(os.path.abspath(fsnative(".")))
    opts, args = opts.parse(argv[1:])
    args[0] = os.path.realpath(args[0])

    app.name = "Ex Falso"
    app.description = _("Audio metadata editor")
    app.id = "io.github.quodlibet.ExFalso"
    app.process_name = "exfalso"
    quodlibet.set_application_info(app)

    import quodlibet.library
    import quodlibet.player

    app.library = quodlibet.library.init()
    app.player = quodlibet.player.init_player("nullbe", app.librarian)
    from quodlibet.qltk.songlist import PlaylistModel

    app.player.setup(PlaylistModel(), None, 0)
    pm = quodlibet.init_plugins()
    pm.rescan()

    from quodlibet.qltk.exfalsowindow import ExFalsoWindow

    dir_ = args[0]
    app.window = ExFalsoWindow(app.library, dir_)
    app.window.init_plugins()

    from quodlibet.util.cover import CoverManager

    app.cover_manager = CoverManager()
    app.cover_manager.init_plugins()

    from quodlibet import session

    session_client = session.init(app)

    quodlibet.enable_periodic_save(save_library=False)
    quodlibet.run(app.window)
    quodlibet.finish_first_session("exfalso")
    config.save()

    session_client.close()

    util.print_d("Finished shutdown.")



# Copyright 2005 Joe Wreschnig
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os

import quodlibet
from quodlibet import _
from quodlibet import util
from quodlibet.qltk import Icons
from quodlibet.plugins.events import EventPlugin

outfile = os.path.join(quodlibet.get_user_dir(), "jabber")
format = """\
<tune xmlns='http://jabber.org/protocol/tune'>
 <artist>%s</artist>
 <title>%s</title>
 <source>%s</source>
 <track>%d</track>
 <length>%d</length>
</tune>"""


class JEP118(EventPlugin):
    PLUGIN_ID = "JEP-118"
    PLUGIN_NAME = _("JEP-118")
    PLUGIN_DESC_MARKUP = _("Outputs a Jabber User Tunes file to %(path)s.") % {
        "path": util.monospace("~/.quodlibet/jabber")
    }
    PLUGIN_ICON = Icons.DOCUMENT_SAVE

    def plugin_on_song_started(self, song):
        if song is None:
            try:
                with open(outfile, "w", encoding="utf-8") as f:
                    f.write("<tune xmlns='http://jabber.org/protocol/tune'/>")
            except OSError:
                pass
        else:
            try:
                with open(outfile, "w", encoding="utf-8") as f:
                    f.write(
                        format
                        % (
                            util.escape(song.comma("artist")),
                            util.escape(song.comma("title")),
                            util.escape(song.comma("album")),
                            song("~#track", 0),
                            song.get("~#length", 0),
                        )
                    )
            except OSError:
                pass



# Copyright 2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gdk

from quodlibet import _
from quodlibet import app
from quodlibet.plugins.events import EventPlugin
from quodlibet.qltk import Icons


class ToggleMenuBarPlugin(EventPlugin):
    PLUGIN_ID = "ToggleMenuBar"
    PLUGIN_NAME = _("Toggle Menu Bar")
    PLUGIN_DESC = _("Toggle the menu bar by pressing the Alt key.")
    PLUGIN_ICON = Icons.EDIT

    def enabled(self):
        window = app.window

        # Maybe this should be made directly accessible
        self._menubar = window.get_children()[0].get_children()[0]

        # Initially set hidden
        self._menubar.set_visible(False)

        # Menu bar visibility toggle
        def toggle_menubar(widget, event):
            if event.keyval in (Gdk.KEY_Alt_L, Gdk.KEY_Alt_R):
                self._menubar.set_visible(not self._menubar.get_visible())
                # Select the menu bar if visible
                if self._menubar.get_visible():
                    self._menubar.select_first(False)
                    self._menubar.deselect()

        self._key_release_handler = window.connect("key_release_event", toggle_menubar)

        # Menu bar hide after deactivation
        def hide_menubar(widget):
            self._menubar.set_visible(False)

        self._deactivate_handler = self._menubar.connect("deactivate", hide_menubar)

    def disabled(self):
        window = app.window
        window.disconnect(self._key_release_handler)
        del self._key_release_handler
        self._menubar.disconnect(self._deactivate_handler)
        self._menubar.set_visible(True)
        del self._deactivate_handler



# Copyright (C) 2012-13 Thomas Vogt
# Copyright (C) 2012-17 Nick Boultbee
# Copyright (C) 2008 Andreas Bombe
# Copyright (C) 2005  Michael Urman
# Based on osd.py (C) 2005 Ton van den Heuvel, Joe Wreshnig
#                 (C) 2004 Gustavo J. A. M. Carneiro
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from collections import namedtuple
from math import pi

import gi

gi.require_version("PangoCairo", "1.0")

from gi.repository import Gtk, GObject, GLib
from gi.repository import Gdk
from gi.repository import Pango, PangoCairo
import cairo

from quodlibet.qltk.image import get_surface_for_pixbuf, get_surface_extents
from quodlibet import qltk
from quodlibet import app
from quodlibet import pattern


class OSDWindow(Gtk.Window):
    __gsignals__ = {
        "fade-finished": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    MARGIN = 50
    """never any closer to the screen edge than this"""

    BORDER = 20
    """text/cover this far apart, from edge"""

    FADETIME = 0.3
    """take this many seconds to fade in or out"""

    MS = 40
    """wait this many milliseconds between steps"""

    def __init__(self, conf, song):
        Gtk.Window.__init__(self, type=Gtk.WindowType.POPUP)
        self.set_type_hint(Gdk.WindowTypeHint.NOTIFICATION)

        screen = self.get_screen()
        rgba = screen.get_rgba_visual()
        if rgba is not None:
            self.set_visual(rgba)

        self.conf = conf
        self.iteration_source = None
        self.fading_in = False
        self.fade_start_time = 0

        mgeo = screen.get_monitor_geometry(conf.monitor)
        textwidth = mgeo.width - 2 * (self.BORDER + self.MARGIN)

        scale_factor = self.get_scale_factor()
        cover_pixbuf = app.cover_manager.get_pixbuf(
            song, conf.coversize * scale_factor, conf.coversize * scale_factor
        )
        coverheight = 0
        coverwidth = 0
        if cover_pixbuf:
            self.cover_surface = get_surface_for_pixbuf(self, cover_pixbuf)
            coverwidth = cover_pixbuf.get_width() // scale_factor
            coverheight = cover_pixbuf.get_height() // scale_factor
            textwidth -= coverwidth + self.BORDER
        else:
            self.cover_surface = None

        layout = self.create_pango_layout("")
        layout.set_alignment(
            (Pango.Alignment.LEFT, Pango.Alignment.CENTER, Pango.Alignment.RIGHT)[
                conf.align
            ]
        )
        layout.set_spacing(Pango.SCALE * 7)
        layout.set_font_description(Pango.FontDescription(conf.font))
        try:
            layout.set_markup(pattern.XMLFromMarkupPattern(conf.string) % song)
        except pattern.Error:
            layout.set_markup("")
        layout.set_width(Pango.SCALE * textwidth)
        layoutsize = layout.get_pixel_size()
        if layoutsize[0] < textwidth:
            layout.set_width(Pango.SCALE * layoutsize[0])
            layoutsize = layout.get_pixel_size()
        self.title_layout = layout

        winw = layoutsize[0] + 2 * self.BORDER
        if coverwidth:
            winw += coverwidth + self.BORDER
        winh = max(coverheight, layoutsize[1]) + 2 * self.BORDER
        self.set_default_size(winw, winh)

        rect = namedtuple("Rect", ["x", "y", "width", "height"])
        rect.x = self.BORDER
        rect.y = (winh - coverheight) // 2
        rect.width = coverwidth
        rect.height = coverheight

        self.cover_rectangle = rect

        winx = int((mgeo.width - winw) * conf.pos_x)
        winx = max(self.MARGIN, min(mgeo.width - self.MARGIN - winw, winx))
        winy = int((mgeo.height - winh) * conf.pos_y)
        winy = max(self.MARGIN, min(mgeo.height - self.MARGIN - winh, winy))
        self.move(winx + mgeo.x, winy + mgeo.y)

    def do_draw(self, cr):
        if self.is_composited():
            self.draw_title_info(cr)
        else:
            # manual transparency rendering follows
            walloc = self.get_allocation()
            wpos = self.get_position()

            if not getattr(self, "_bg_sf", None):
                # copy the root surface into a temp image surface
                root_win = self.get_root_window()
                bg_sf = cairo.ImageSurface(
                    cairo.FORMAT_ARGB32, walloc.width, walloc.height
                )
                pb = Gdk.pixbuf_get_from_window(
                    root_win, wpos[0], wpos[1], walloc.width, walloc.height
                )
                bg_cr = cairo.Context(bg_sf)
                Gdk.cairo_set_source_pixbuf(bg_cr, pb, 0, 0)
                bg_cr.paint()
                self._bg_sf = bg_sf

            if not getattr(self, "_fg_sf", None):
                # draw the window content in another temp surface
                fg_sf = cairo.ImageSurface(
                    cairo.FORMAT_ARGB32, walloc.width, walloc.height
                )
                fg_cr = cairo.Context(fg_sf)
                fg_cr.set_source_surface(fg_sf)
                self.draw_title_info(fg_cr)
                self._fg_sf = fg_sf

            # first draw the background so we have 'transparancy'
            cr.set_operator(cairo.OPERATOR_SOURCE)
            cr.set_source_surface(self._bg_sf)
            cr.paint()

            # then draw the window content with the right opacity
            cr.set_operator(cairo.OPERATOR_OVER)
            cr.set_source_surface(self._fg_sf)
            cr.paint_with_alpha(self.get_opacity())

    @staticmethod
    def rounded_rectangle(cr, x, y, radius, width, height):
        cr.move_to(x + radius, y)
        cr.line_to(x + width - radius, y)
        cr.arc(
            x + width - radius, y + radius, radius, -90.0 * pi / 180.0, 0.0 * pi / 180.0
        )
        cr.line_to(x + width, y + height - radius)
        cr.arc(
            x + width - radius,
            y + height - radius,
            radius,
            0.0 * pi / 180.0,
            90.0 * pi / 180.0,
        )
        cr.line_to(x + radius, y + height)
        cr.arc(
            x + radius,
            y + height - radius,
            radius,
            90.0 * pi / 180.0,
            180.0 * pi / 180.0,
        )
        cr.line_to(x, y + radius)
        cr.arc(x + radius, y + radius, radius, 180.0 * pi / 180.0, 270.0 * pi / 180.0)
        cr.close_path()

    @property
    def corners_factor(self):
        if self.conf.corners != 0:
            return 0.14
        return 0.0

    def draw_conf_rect(self, cr, x, y, width, height, radius):
        if self.conf.corners != 0:
            self.rounded_rectangle(cr, x, y, radius, width, height)
        else:
            cr.rectangle(x, y, width, height)

    def draw_title_info(self, cr):
        cr.save()
        do_shadow = self.conf.shadow[0] != -1.0
        do_outline = self.conf.outline[0] != -1.0

        self.set_name("osd_bubble")
        qltk.add_css(
            self,
            """
            #osd_bubble {
                background-color:rgba(0,0,0,0);
            }
        """,
        )

        cr.set_operator(cairo.OPERATOR_OVER)
        cr.set_source_rgba(*self.conf.fill)
        radius = min(25, self.corners_factor * min(*self.get_size()))
        self.draw_conf_rect(cr, 0, 0, self.get_size()[0], self.get_size()[1], radius)
        cr.fill()

        # draw border
        if do_outline:
            # Make border darker and more translucent than the fill
            f = self.conf.fill
            rgba = (f[0] / 1.25, f[1] / 1.25, f[2] / 1.25, f[3] / 2.0)
            cr.set_source_rgba(*rgba)
            self.draw_conf_rect(
                cr, 1, 1, self.get_size()[0] - 2, self.get_size()[1] - 2, radius
            )
            cr.set_line_width(2.0)
            cr.stroke()

        textx = self.BORDER

        if self.cover_surface is not None:
            rect = self.cover_rectangle
            textx += rect.width + self.BORDER
            surface = self.cover_surface
            transmat = cairo.Matrix()

            if do_shadow:
                cr.set_source_rgba(*self.conf.shadow)
                self.draw_conf_rect(
                    cr,
                    rect.x + 2,
                    rect.y + 2,
                    rect.width,
                    rect.height,
                    0.6 * self.corners_factor * rect.width,
                )
                cr.fill()

            if do_outline:
                cr.set_source_rgba(*self.conf.outline)
                self.draw_conf_rect(
                    cr,
                    rect.x,
                    rect.y,
                    rect.width,
                    rect.height,
                    0.6 * self.corners_factor * rect.width,
                )
                cr.stroke()

            cr.set_source_surface(surface, 0, 0)
            width, height = get_surface_extents(surface)[2:]

            transmat.scale(width / float(rect.width), height / float(rect.height))
            transmat.translate(-rect.x, -rect.y)
            cr.get_source().set_matrix(transmat)
            self.draw_conf_rect(
                cr,
                rect.x,
                rect.y,
                rect.width,
                rect.height,
                0.6 * self.corners_factor * rect.width,
            )
            cr.fill()

        PangoCairo.update_layout(cr, self.title_layout)
        height = self.title_layout.get_pixel_size()[1]
        texty = (self.get_size()[1] - height) // 2

        if do_shadow:
            cr.set_source_rgba(*self.conf.shadow)
            cr.move_to(textx + 2, texty + 2)
            PangoCairo.show_layout(cr, self.title_layout)
        if do_outline:
            cr.set_source_rgba(*self.conf.outline)
            cr.move_to(textx, texty)
            PangoCairo.layout_path(cr, self.title_layout)
            cr.stroke()
        cr.set_source_rgb(*self.conf.text[:3])
        cr.move_to(textx, texty)
        PangoCairo.show_layout(cr, self.title_layout)
        cr.restore()

    def fade_in(self):
        self.do_fade_inout(True)

    def fade_out(self):
        self.do_fade_inout(False)

    def do_fade_inout(self, fadein):
        fadein = bool(fadein)
        self.fading_in = fadein
        now = GLib.get_real_time()

        fraction = self.get_opacity()
        if not fadein:
            fraction = 1.0 - fraction
        self.fade_start_time = now - fraction * self.FADETIME

        if self.iteration_source is None:
            self.iteration_source = GLib.timeout_add(
                self.MS, self.fade_iteration_callback
            )

    def fade_iteration_callback(self):
        delta = GLib.get_real_time() - self.fade_start_time
        fraction = delta / self.FADETIME

        if self.fading_in:
            self.set_opacity(fraction)
        else:
            self.set_opacity(1.0 - fraction)

        if not self.is_composited():
            self.queue_draw()

        if fraction >= 1.0:
            self.iteration_source = None
            self.emit("fade-finished", self.fading_in)
            return False
        return True



# Copyright 2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


from os.path import splitext
from pathlib import Path
from time import sleep
from typing import Any
from collections.abc import Collection
from urllib.parse import urlparse

from gi.repository import Soup, GObject

from quodlibet import print_d, print_w, _, print_e
from quodlibet.formats import AudioFile
from quodlibet.qltk.notif import Task
from quodlibet.util import http, format_size
from quodlibet.util.path import escape_filename


class DownloadProgress(GObject.Object):
    """Downloads songs asynchronously, updating a Task"""

    __gsignals__ = {
        "finished": (GObject.SignalFlags.RUN_LAST, None, (object, object)),
    }

    def __init__(self, songs: Collection[AudioFile], task=None) -> None:
        super().__init__()
        self.songs = songs
        self.successful: set[AudioFile] = set()
        self.failed: set[AudioFile] = set()
        self.task = task or Task(_("Browser"), _("Downloading files"))

    def success(self, song: AudioFile) -> None:
        self.successful.add(song)
        self._update()

    def failure(self, song: AudioFile) -> None:
        self.failed.add(song)
        self._update()

    def _update(self) -> None:
        frac = self.frac
        print_d(f"At {frac * 100:.0f}% ({len(self.successful)}, {len(self.failed)})")
        self.task.update(frac)

    @property
    def frac(self):
        return (len(self.successful) + len(self.failed)) / len(self.songs)

    def _downloaded(self, msg: Soup.Message, result: Any, data: tuple) -> None:
        path, song = data
        try:
            headers = msg.get_property("response-headers")
            size = int(headers.get("content-length"))
            content_type = headers.get("content-type")
            print_d(
                f"Downloaded {format_size(size)} of {content_type}: {song('title')}"
            )
            _, ext = splitext(urlparse(song("~uri")).path)
            fn = (
                escape_filename(song("~artist~title")[:100], safe=b" ,';")
                or song("~basename")
                or f"download-{hash(song('~filename'))}"
            )
            path = path / Path(fn + ext)
            if path.is_file() and path.stat():
                print_w(f"{path!s} already exists. Skipping download")
                self.success(song)
                return
            with open(path, "wb") as f:
                f.write(result)
            self.success(song)
            print_d(f"Downloaded to {path} successfully!")
        except Exception as e:
            print_e(f"Failed download ({e})")
            self.failure(song)

    def _failed(self, _req: Any, _exc: Exception, data: tuple) -> None:
        path, song = data
        self.failure(song)

    def download_songs(self, path: Path):
        for s in self.songs:
            msg = Soup.Message.new("GET", s("~uri"))
            http.download(
                msg,
                cancellable=None,
                callback=self._downloaded,
                failure_callback=self._failed,
                data=(path, s),
            )
            yield
        while self.frac < 1 and self.task:
            sleep(0.1)
            yield
        self.task.finish()
        self.emit("finished", len(self.successful), len(self.failed))



# Copyright 2013,2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import sys
import os
import ctypes


if sys.platform == "win32":
    from . import winapi
    from .winapi import (
        SHGFPType,
        CSIDLFlag,
        CSIDL,
        GUID,
        SHGetFolderPathW,
        S_OK,
        MAX_PATH,
        KnownFolderFlag,
        FOLDERID,
        SHGetKnownFolderPath,
        CoTaskMemFree,
        CoInitialize,
        IShellLinkW,
        CoCreateInstance,
        CLSID_ShellLink,
        CLSCTX_INPROC_SERVER,
        IPersistFile,
    )


def open_folder_and_select_items(folder, items=None):
    """Shows a directory and optional files or subdirectories in the
    file manager (explorer.exe).

    If both folder and items is given the file manager will
    display the content of `folder` and highlight all `items`.

    If only a directory is given then the content of the parent directory is
    shown and the `folder` highlighted.

    Might raise WindowsError in case something fails (any of the
    files not existing etc.)
    """

    if items is None:
        items = []

    assert isinstance(folder, str)
    for item in items:
        assert isinstance(item, str)
        assert not os.path.split(item)[0]

    desktop = winapi.IShellFolder()
    parent = winapi.IShellFolder()
    parent_id = winapi.PIDLIST_ABSOLUTE()
    child_ids = (winapi.PIDLIST_RELATIVE * len(items))()

    try:
        winapi.CoInitialize(None)

        winapi.SHParseDisplayName(folder, None, ctypes.byref(parent_id), 0, None)

        winapi.SHGetDesktopFolder(ctypes.byref(desktop))

        desktop.BindToObject(
            parent_id, None, winapi.IShellFolder.IID, ctypes.byref(parent)
        )

        for i, item in enumerate(items):
            attrs = winapi.ULONG(0)
            parent.ParseDisplayName(
                None, None, item, None, ctypes.byref(child_ids[i]), ctypes.byref(attrs)
            )

        winapi.SHOpenFolderAndSelectItems(
            parent_id, len(child_ids), winapi.PCUITEMID_CHILD_ARRAY(child_ids), 0
        )
    finally:
        for child_id in child_ids:
            if child_id:
                winapi.CoTaskMemFree(child_id)
        if parent_id:
            winapi.ILFree(parent_id)
        if parent:
            parent.Release()
        if desktop:
            desktop.Release()


def _get_path(folder, default=False, create=False):
    """A path to an directory or None.

    Takes a CSIDL instance as folder.
    """

    if default:
        flags = SHGFPType.DEFAULT
    else:
        flags = SHGFPType.CURRENT

    if create:
        folder |= CSIDLFlag.CREATE

    # we don't want env vars
    folder |= CSIDLFlag.DONT_UNEXPAND

    buffer_ = ctypes.create_unicode_buffer(MAX_PATH)
    try:
        result = SHGetFolderPathW(0, folder, 0, flags, buffer_)
    except OSError:
        return None
    if result != S_OK:
        return None
    return buffer_.value


def _get_known_path(folder, default=False, create=False):
    """A path to an directory or None

    Takes a FOLDERID instances as folder.
    """

    if default:
        flags = KnownFolderFlag.DEFAULT_PATH
    else:
        flags = 0

    if create:
        flags |= KnownFolderFlag.CREATE

    flags |= KnownFolderFlag.DONT_VERIFY

    ptr = ctypes.c_wchar_p()
    guid = GUID(folder)
    try:
        result = SHGetKnownFolderPath(
            ctypes.byref(guid), flags, None, ctypes.byref(ptr)
        )
    except OSError:
        return None
    if result != S_OK:
        return None
    path = ptr.value
    CoTaskMemFree(ptr)
    return path


def get_personal_dir(**kwargs):
    r"""e.g. 'C:\Users\<user>\Documents'"""

    return _get_path(CSIDL.PERSONAL, **kwargs)


def get_appdata_dir(**kwargs):
    r"""e.g. 'C:\Users\<user>\AppData\Roaming'"""

    return _get_path(CSIDL.APPDATA, **kwargs)


def get_desktop_dir(**kwargs):
    r"""e.g. 'C:\Users\<user>\Desktop'"""

    return _get_path(CSIDL.DESKTOP, **kwargs)


def get_music_dir(**kwargs):
    r"""e.g. 'C:\Users\<user>\Music'"""

    return _get_path(CSIDL.MYMUSIC, **kwargs)


def get_profile_dir(**kwargs):
    r"""e.g. 'C:\Users\<user>'"""

    return _get_path(CSIDL.PROFILE, **kwargs)


def get_links_dir(**kwargs):
    r"""e.g. 'C:\Users\<user>\Links'"""

    return _get_known_path(FOLDERID.LINKS, **kwargs)


def get_link_target(path):
    """Takes a path to a .lnk file and returns a path the .lnk file
    is targeting.

    Might raise WindowsError in case something fails.
    """

    assert isinstance(path, str)

    CoInitialize(None)

    pShellLinkW = IShellLinkW()
    CoCreateInstance(
        ctypes.byref(CLSID_ShellLink),
        None,
        CLSCTX_INPROC_SERVER,
        ctypes.byref(IShellLinkW.IID),
        ctypes.byref(pShellLinkW),
    )
    try:
        pPersistFile = IPersistFile()
        pShellLinkW.QueryInterface(
            ctypes.byref(IPersistFile.IID), ctypes.byref(pPersistFile)
        )
        try:
            buffer_ = ctypes.create_unicode_buffer(path, MAX_PATH)
            pPersistFile.Load(buffer_, 0)
        finally:
            pPersistFile.Release()
        pShellLinkW.GetPath(buffer_, MAX_PATH, None, 0)
    finally:
        pShellLinkW.Release()

    return ctypes.wstring_at(buffer_)



# Copyright 2005 Joe Wreschnig, Michael Urman
#        2012-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import config


class ConfigCheckButton(Gtk.CheckButton):
    """A CheckButton that connects to QL's config module, and toggles
    a boolean configuration value when it is toggled.

    It is initialised to the current config value if `populate` is set True."""

    def __init__(
        self, label, section, option, populate=False, tooltip=None, default=None
    ):
        super().__init__(label=label, use_underline=True)

        if default is None:
            default = config._config.defaults.getboolean(section, option, True)

        if populate:
            self.set_active(config.getboolean(section, option, default))
        if tooltip:
            self.set_tooltip_text(tooltip)
        self.connect("toggled", ConfigCheckButton.__toggled, section, option)

    def __toggled(self, section, option):
        config.set(section, option, str(bool(self.get_active())).lower())


class ConfigSwitch(Gtk.Box):
    """A Switch that connects to QL's config module, and toggles
    a boolean configuration value when it is toggled.

    It is initialised to the current config value if `populate` is set True."""

    def __init__(
        self, label, section, option, populate=False, tooltip=None, default=None
    ):
        super().__init__()
        self.switch = Gtk.Switch()
        eb = Gtk.EventBox()
        if label is not None:
            self.label = Gtk.Label(label, use_underline=True)
            self.label.set_mnemonic_widget(self.switch)
            eb.add(self.label)
        self.pack_start(eb, False, True, 0)
        self.pack_end(self.switch, False, True, 0)
        if default is None:
            default = config._config.defaults.getboolean(section, option, True)

        if populate:
            self.set_active(config.getboolean(section, option, default))
        if tooltip:
            self.switch.set_tooltip_text(tooltip)
        self.switch.connect("notify::active", self.__activated, section, option)
        eb.connect(
            "button_press_event",
            lambda *_: self.switch.set_state(not self.switch.get_state()),
        )

    def set_active(self, value: bool):
        self.switch.set_active(value)

    def get_active(self) -> bool:
        return self.switch.get_active()

    def connect(self, *args, **kwargs):
        self.switch.connect(*args, **kwargs)

    def __activated(self, switch, state, section, option):
        config.set(section, option, str(switch.get_active()).lower())


class ConfigCheckMenuItem(Gtk.CheckMenuItem):
    """A CheckMenuItem that connects to QL's config module, and toggles
    a boolean configuration value when it is toggled.

    It is initialised to the current config value if `populate` is set True."""

    def __init__(self, label, section, option, populate=False, default=False):
        super().__init__(label=label, use_underline=True)
        if populate:
            self.set_active(config.getboolean(section, option, default))
        self.connect("toggled", ConfigCheckMenuItem.__toggled, section, option)

    def __toggled(self, section, option):
        config.set(section, option, str(bool(self.get_active())).lower())



# Copyright 2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gdk


def mix(src: Gdk.RGBA, dest: Gdk.RGBA, ratio: float) -> Gdk.RGBA:
    """Mixes two Gdk colours into a result"""
    ratio = min(1.0, max(0.0, ratio))
    inv = 1.0 - ratio
    return Gdk.RGBA(
        inv * src.red + ratio * dest.red,
        inv * src.green + ratio * dest.green,
        inv * src.blue + ratio * dest.blue,
        inv * src.alpha + ratio * dest.alpha,
    )



# Copyright 2004-2007 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2009-2010 Steven Robertson
#         2012 - 2023 Nick Boultbee
#           2009-2013 Christoph Reiter
#                2016 Mice Pápai
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import config
from quodlibet import qltk
from quodlibet import util
from quodlibet import _
from quodlibet.browsers._base import FakeDisplayItem, EditDisplayPatternMixin
from quodlibet.formats import PEOPLE
from quodlibet.qltk import Button, Icons
from quodlibet.qltk.ccb import ConfigCheckButton
from quodlibet.util import format_rating
from quodlibet.util.i18n import numeric_phrase
from quodlibet.util.dprint import print_d

PEOPLE  # noqa
_SOME_PEOPLE = "\n".join(
    [
        util.tag("artist"),
        util.tag("performer"),
        util.tag("composer"),
        util.tag("arranger"),
    ]
)

_EMPTY = _("Songs not in an album")
DEFAULT_PATTERN_TEXT = f"""[b]<album|<album>|{_EMPTY}>[/b]<date| (<date>)>
[small]<~discs|<~discs> - ><~tracks> - <~long-length>[/small]
<~people>"""


class Preferences(qltk.UniqueWindow, EditDisplayPatternMixin):
    _DEFAULT_PATTERN = DEFAULT_PATTERN_TEXT

    _PREVIEW_ITEM = FakeDisplayItem(
        {
            "date": "2010-10-31",
            "~length": util.format_time_display(6319),
            "~long-length": util.format_time_long(6319),
            "~tracks": numeric_phrase("%d track", "%d tracks", 5),
            "~discs": numeric_phrase("%d disc", "%d discs", 2),
            "~#rating": 0.75,
            "album": _("An Example Album"),
            "~people": _SOME_PEOPLE + "...",
        }
    )

    def __init__(self, browser):
        if self.is_not_unique():
            return
        super().__init__()
        self.set_border_width(12)
        self.set_title(_("Cover Grid Preferences"))
        self.set_default_size(420, 380)
        self.set_transient_for(qltk.get_top_parent(browser))
        # Do this config-driven setup at instance-time
        self._PREVIEW_ITEM["~rating"] = format_rating(0.75)
        self.mag_lock = False

        box = Gtk.VBox(spacing=6)
        vbox = Gtk.VBox(spacing=6)
        cb = ConfigCheckButton(_("Show album _text"), "browsers", "album_text")
        cb.set_active(config.getboolean("browsers", "album_text", True))
        cb.connect("toggled", lambda s: browser.toggle_text())
        vbox.pack_start(cb, False, True, 0)

        cb2 = ConfigCheckButton(
            _('Show "All Albums" Item'), "browsers", "covergrid_all"
        )
        cb2.set_active(config.getboolean("browsers", "covergrid_all", True))
        cb2.connect("toggled", lambda s: browser.toggle_item_all())
        vbox.pack_start(cb2, False, True, 0)

        cb3 = ConfigCheckButton(_("Wide Mode"), "browsers", "covergrid_wide")
        cb3.set_active(config.getboolean("browsers", "covergrid_wide", False))
        cb3.connect("toggled", lambda s: browser.toggle_wide())
        vbox.pack_start(cb3, False, True, 0)

        def mag_changed(mag):
            newmag = mag.get_value()
            oldmag = config.getfloat("browsers", "covergrid_magnification", 3.0)
            if newmag == oldmag:
                print_d(f"Covergrid magnification haven't changed: {newmag}")
                return
            print_d(f"Covergrid magnification update from {oldmag} to {newmag}")
            config.set("browsers", "covergrid_magnification", mag.get_value())
            browser.update_mag()

        mag_scale = Gtk.HScale(
            adjustment=Gtk.Adjustment.new(
                config.getfloat("browsers", "covergrid_magnification", 3),
                1,
                10.0,
                0.5,
                0.5,
                0,
            )
        )
        mag_scale.set_tooltip_text(_("Cover Magnification"))
        l = Gtk.Label(label=_("Cover Magnification"))
        mag_scale.set_value_pos(Gtk.PositionType.RIGHT)
        mag_scale.connect("value-changed", mag_changed)

        vbox.pack_start(l, False, True, 0)
        vbox.pack_start(mag_scale, False, True, 0)

        f = qltk.Frame(_("Options"), child=vbox)
        box.pack_start(f, False, True, 12)

        display_frame = self.edit_display_pane(browser, _("Album Display"))
        box.pack_start(display_frame, True, True, 0)

        main_box = Gtk.VBox(spacing=12)
        close = Button(_("_Close"), Icons.WINDOW_CLOSE)
        close.connect("clicked", lambda *x: self.destroy())
        b = Gtk.HButtonBox()
        b.set_layout(Gtk.ButtonBoxStyle.END)
        b.pack_start(close, True, True, 0)

        main_box.pack_start(box, True, True, 0)
        self.use_header_bar()

        if not self.has_close_button():
            main_box.pack_start(b, False, True, 0)
        self.add(main_box)

        close.grab_focus()
        self.show_all()



# Copyright 2014 Christoph Reiter
#           2017-2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import signal
import stat
from tempfile import mkstemp

try:
    import fcntl
except ImportError:
    pass

from gi.repository import GLib
from senf import fsn2bytes

from quodlibet import print_d, print_e
from quodlibet.util.path import mkdir

FIFO_TIMEOUT = 10
"""time in seconds until we give up writing/reading"""


class FIFOError(Exception):
    pass


def _sigalrm_timeout(*args):
    raise TimeoutError


def _write_fifo(fifo_path, data):
    """Writes the data to the FIFO or raises `FIFOError`"""

    assert isinstance(data, bytes)

    # This will raise if the FIFO doesn't exist or there is no reader
    try:
        fifo = os.open(fifo_path, os.O_WRONLY | os.O_NONBLOCK)
    except OSError:
        try:
            os.unlink(fifo_path)
        except OSError:
            pass
        raise
    else:
        try:
            os.close(fifo)
        except OSError:
            pass

    try:
        signal.signal(signal.SIGALRM, _sigalrm_timeout)
        signal.alarm(FIFO_TIMEOUT)
        with open(fifo_path, "wb") as f:
            signal.signal(signal.SIGALRM, signal.SIG_IGN)
            f.write(data)
    except (OSError, TimeoutError) as e:
        # Unable to write to the fifo. Removing it.
        try:
            os.unlink(fifo_path)
        except OSError:
            pass
        raise FIFOError(f"Couldn't write to fifo {fifo_path!r}") from e


def split_message(data):
    """Split incoming data in pairs of (command, FIFO path or `None`)

    This supports two data formats:
    Newline-separated commands without a return FIFO path.
    and "NULL<command>NULL<fifo-path>NULL"

    Args:
        data (bytes)
    Returns:
        Tuple[bytes, bytes]
    Raises:
        ValueError
    """

    assert isinstance(data, bytes)

    while data:
        elem, null, data = data.partition(b"\x00")
        if elem:  # newline-separated commands
            yield from ((line, None) for line in elem.splitlines())
            data = null + data
        else:  # NULL<command>NULL<fifo-path>NULL
            cmd, fifo_path, data = data.split(b"\x00", 2)
            yield (cmd, fifo_path)


def write_fifo(fifo_path, data):
    """Writes the data to the FIFO and returns a response.

    Args:
        fifo_path (pathlike)
        data (bytes)
    Returns:
        bytes
    Raises:
        FIFOError
    """

    assert isinstance(data, bytes)

    fd, filename = mkstemp()
    try:
        os.close(fd)
        os.unlink(filename)
        # mkfifo fails if the file exists, so this is safe.
        os.mkfifo(filename, 0o600)

        _write_fifo(
            fifo_path, b"\x00".join([b"", data, fsn2bytes(filename, None), b""])
        )

        try:
            signal.signal(signal.SIGALRM, _sigalrm_timeout)
            signal.alarm(FIFO_TIMEOUT)
            with open(filename, "rb") as h:
                signal.signal(signal.SIGALRM, signal.SIG_IGN)
                return h.read()
        except TimeoutError as e:
            # In case the main instance deadlocks we can write to it, but
            # reading will time out. Assume it is broken and delete the
            # fifo.
            try:
                os.unlink(fifo_path)
            except OSError:
                pass
            raise FIFOError("Timeout reached") from e
    except OSError as e:
        raise FIFOError(*e.args) from e
    finally:
        try:
            os.unlink(filename)
        except OSError:
            pass


def fifo_exists(fifo_path):
    """Returns whether a FIFO exists (and is writeable).

    Args:
        fifo_path (pathlike)
    Returns:
        bool
    """

    # https://github.com/quodlibet/quodlibet/issues/1131
    # FIXME: There is a race where control() creates a new file
    # instead of writing to the FIFO, confusing the next QL instance.
    # Remove non-FIFOs here for now.
    try:
        if not stat.S_ISFIFO(os.stat(fifo_path).st_mode):
            print_d(f"{fifo_path!r} not a FIFO. Removing it.")
            os.remove(fifo_path)
    except OSError:
        pass
    return os.path.exists(fifo_path)


class FIFO:
    """Creates and reads from a FIFO"""

    _source_id = None

    def __init__(self, path, callback):
        """
        Args:
            path (pathlike)
            callback (Callable[[bytes], None])
        """

        self._callback = callback
        self._path = path

    def destroy(self):
        """After destroy() the callback will no longer be called
        and the FIFO can no longer be used. Can be called multiple
        times.
        """

        if self._source_id is not None:
            GLib.source_remove(self._source_id)
            self._source_id = None

        try:
            os.unlink(self._path)
        except OSError:
            pass

    def open(self, ignore_lock=False):
        """Create the FIFO and listen to it.

        Raises:
            FIFOError in case another process is already using it.
        """
        from quodlibet import qltk

        mkdir(os.path.dirname(self._path))
        try:
            os.mkfifo(self._path, 0o600)
        except OSError:
            # maybe exists, we'll fail below otherwise
            pass

        try:
            fifo = os.open(self._path, os.O_NONBLOCK)
        except OSError:
            return

        while True:
            try:
                fcntl.flock(fifo, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except InterruptedError:  # EINTR
                continue
            except BlockingIOError as e:  # EWOULDBLOCK
                if not ignore_lock:
                    raise FIFOError("fifo already locked") from e
            except OSError as e:
                print_d(f"fifo locking failed: {e!r}")
            break

        try:
            f = os.fdopen(fifo, "rb", 4096)
        except OSError as e:
            print_e(f"Couldn't open FIFO ({e!r})")
        else:
            self._source_id = qltk.io_add_watch(
                f,
                GLib.PRIORITY_DEFAULT,
                GLib.IO_IN | GLib.IO_ERR | GLib.IO_HUP,
                self._process,
            )

    def _process(self, source, condition):
        if condition in {GLib.IO_ERR, GLib.IO_HUP}:
            self.open(ignore_lock=True)
            return False

        while True:
            try:
                data = source.read()
            except InterruptedError:  # EINTR
                continue
            except (BlockingIOError, PermissionError):  # EWOULDBLOCK, EACCES
                return True
            except OSError:
                self.open(ignore_lock=True)
                return False
            break

        if not data:
            self.open(ignore_lock=True)
            return False

        self._callback(data)

        return True



# Copyright 2006-2007 Lukas Lalinsky
#                2021 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import GLib

from quodlibet import _
from quodlibet import config
from quodlibet.player import PlayerError
from quodlibet.player._base import BasePlayer
from quodlibet.util.string import decode

from . import cdefs
from .cdefs import (
    XINE_PARAM_SPEED,
    XINE_PARAM_GAPLESS_SWITCH,
    xine_dispose,
    XINE_SPEED_PAUSE,
    xine_play,
    xine_close,
    xine_set_param,
    xine_get_param,
    xine_get_status,
    xine_open,
    xine_stop,
    XINE_PARAM_EARLY_FINISHED_EVENT,
    XINE_META_INFO_ARTIST,
    xine_ui_message_data_t,
    XINE_EVENT_UI_SET_TITLE,
    XINE_PARAM_AUDIO_AMP_MUTE,
    XINE_PARAM_AUDIO_AMP_LEVEL,
    xine_new,
    XINE_EVENT_UI_PLAYBACK_FINISHED,
    xine_event_dispose_queue,
    xine_init,
    XINE_PARAM_IGNORE_VIDEO,
    XINE_PARAM_IGNORE_SPU,
    xine_config_load,
    xine_check_version,
    xine_get_homedir,
    xine_list_input_plugins,
    xine_exit,
    xine_open_audio_driver,
    xine_close_audio_driver,
    XINE_STATUS_PLAY,
    XINE_SPEED_NORMAL,
    xine_get_pos_length,
    XINE_MSG_NO_ERROR,
    XINE_EVENT_UI_MESSAGE,
    xine_get_meta_info,
    XINE_META_INFO_ALBUM,
    XINE_META_INFO_TITLE,
    xine_stream_new,
    xine_get_version_string,
    xine_event_new_queue,
    xine_event_create_listener_thread,
)


class XineHandle:
    def __init__(self):
        _xine = xine_new()
        xine_config_load(_xine, xine_get_homedir() + b"/.xine/config")
        xine_init(_xine)
        self._xine = _xine

    def list_input_plugins(self):
        """
        Returns:
            List[str]
        """

        plugins = []
        for plugin in xine_list_input_plugins(self._xine):
            if not plugin:
                break
            plugins.append(decode(plugin))
        return plugins

    def exit(self):
        xine_exit(self._xine)

    def open_audio_driver(self, identifier, data):
        return xine_open_audio_driver(self._xine, identifier, data)

    def close_audio_driver(self, driver):
        xine_close_audio_driver(self._xine, driver)

    def stream_new(self, audio_port, video_port):
        return xine_stream_new(self._xine, audio_port, video_port)


class XinePlaylistPlayer(BasePlayer):
    """Xine playlist player."""

    _paused = True

    def __init__(self, driver, librarian):
        """May raise PlayerError"""

        super().__init__()
        self.name = "xine"
        self.version_info = "xine-lib: " + decode(xine_get_version_string())
        self._volume = 1.0
        self._handle = XineHandle()
        self._supports_gapless = xine_check_version(1, 1, 1) == 1
        self._event_queue = None

        self._new_stream(driver)
        self._librarian = librarian
        self._destroyed = False

    def _new_stream(self, driver):
        assert driver is None or isinstance(driver, bytes)
        self._audio_port = self._handle.open_audio_driver(driver, None)
        if not self._audio_port:
            raise PlayerError(
                _("Unable to create audio output"),
                _(
                    "The audio device %r was not found. Check your Xine "
                    "settings in ~/.quodlibet/config."
                )
                % driver,
            )
        self._stream = self._handle.stream_new(self._audio_port, None)
        xine_set_param(self._stream, XINE_PARAM_IGNORE_VIDEO, 1)
        xine_set_param(self._stream, XINE_PARAM_IGNORE_SPU, 1)
        self.update_eq_values()
        if self._supports_gapless:
            xine_set_param(self._stream, XINE_PARAM_EARLY_FINISHED_EVENT, 1)
        if self._event_queue:
            xine_event_dispose_queue(self._event_queue)
        self._event_queue = xine_event_new_queue(self._stream)
        xine_event_create_listener_thread(self._event_queue, self._event_listener, None)

    def _destroy(self):
        self._destroyed = True

        if self._stream:
            xine_close(self._stream)
            xine_dispose(self._stream)
            self._stream = None
        if self._event_queue:
            xine_event_dispose_queue(self._event_queue)
        if self._audio_port:
            self._handle.close_audio_driver(self._audio_port)
        self._handle.exit()

    def _playback_finished(self):
        if self._destroyed:
            return False

        self._source.next_ended()
        self._end(False, None, gapless=True)
        return False

    def _update_metadata(self):
        if self._destroyed:
            return False

        if not self.song or not self.song.multisong:
            return False
        if self.info is self.song:
            self.info = type(self.song)(self.song["~filename"])
            self.info.multisong = False
        changed = False
        meta = [
            (XINE_META_INFO_TITLE, "title"),
            (XINE_META_INFO_ARTIST, "artist"),
            (XINE_META_INFO_ALBUM, "album"),
        ]
        for info, name in meta:
            text = xine_get_meta_info(self._stream, info)
            if not text:
                continue
            text = text.decode("UTF-8", "replace")
            if self.info.get(name) != text:
                self.info[name] = text
                changed = True
        if changed:
            self.emit("song-started", self.info)
            if self._librarian is not None:
                self._librarian.changed([self.song])
        return False

    def _event_listener(self, user_data, event):
        event = event.contents
        if event.type == XINE_EVENT_UI_PLAYBACK_FINISHED:
            GLib.idle_add(self._playback_finished, priority=GLib.PRIORITY_HIGH)
        elif event.type == XINE_EVENT_UI_SET_TITLE:
            GLib.idle_add(self._update_metadata, priority=GLib.PRIORITY_HIGH)
        elif event.type == XINE_EVENT_UI_MESSAGE:
            from ctypes import POINTER, cast, string_at, addressof

            msg = cast(event.data, POINTER(xine_ui_message_data_t)).contents
            if msg.type != XINE_MSG_NO_ERROR:
                if msg.explanation:
                    message = string_at(addressof(msg) + msg.explanation)
                else:
                    message = f"xine error {msg.type}"
                message = message.decode("utf-8", errors="replace")
                GLib.idle_add(self._error, PlayerError(message))
        return True

    def do_get_property(self, property):
        if property.name == "volume":
            return self._volume
        if property.name == "seekable":
            if self.song is None:
                return False
            return True
        if property.name == "mute":
            if not self._destroyed:
                return xine_get_param(self._stream, XINE_PARAM_AUDIO_AMP_MUTE)
            return False
        raise AttributeError

    def do_set_property(self, property, v):
        if property.name == "volume":
            self._volume = v
            v = self.calc_replaygain_volume(v)
            v = min(100, int(v * 100))
            if not self._destroyed:
                xine_set_param(self._stream, XINE_PARAM_AUDIO_AMP_LEVEL, v)
        elif property.name == "mute":
            if not self._destroyed:
                xine_set_param(self._stream, XINE_PARAM_AUDIO_AMP_MUTE, v)
        else:
            raise AttributeError

    def get_position(self):
        """Return the current playback position in milliseconds,
        or 0 if no song is playing."""
        pos_stream, pos_time, length_time = xine_get_pos_length(self._stream)
        return pos_time

    def _stop(self):
        xine_stop(self._stream)

    def _pause(self):
        xine_set_param(self._stream, XINE_PARAM_SPEED, XINE_SPEED_PAUSE)

    def _play(self):
        if xine_get_param(self._stream, XINE_PARAM_SPEED) != XINE_SPEED_NORMAL:
            xine_set_param(self._stream, XINE_PARAM_SPEED, XINE_SPEED_NORMAL)
        if xine_get_status(self._stream) != XINE_STATUS_PLAY:
            xine_play(self._stream, 0, 0)

    @property
    def paused(self):
        return self._paused

    @paused.setter
    def paused(self, paused):
        if paused == self._paused:
            return
        self._paused = paused
        self.emit((paused and "paused") or "unpaused")
        if self._paused != paused:
            return

        if self.song:
            if paused:
                if not self.song.is_file:
                    xine_close(self._stream)
                    xine_open(self._stream, self.song("~uri").encode("ascii"))
                else:
                    self._pause()
            else:
                self._play()

    def _error(self, player_error=None):
        if self._destroyed:
            return False

        if self.error:
            return False

        self.error = True
        self.paused = True
        if player_error:
            self.emit("error", self.song, player_error)
        return None

    def seek(self, pos):
        """Seek to a position in the song, in milliseconds."""
        if xine_get_param(self._stream, XINE_PARAM_SPEED) == XINE_SPEED_PAUSE:
            xine_play(self._stream, 0, int(pos))
            xine_set_param(self._stream, XINE_PARAM_SPEED, XINE_SPEED_PAUSE)
        else:
            xine_play(self._stream, 0, int(pos))
        self.emit("seek", self.song, pos)

    def _end(self, stopped, next_song=None, gapless=False):
        # We need to set self.song to None before calling our signal
        # handlers. Otherwise, if they try to end the song they're given
        # (e.g. by removing it), then we get in an infinite loop.
        song = self.song
        self.song = self.info = None
        self.emit("song-ended", song, stopped)

        # reset error state
        self.error = False

        current = next_song if next_song else (self._source and self._source.current)

        # Then, set up the next song.
        self.song = self.info = current
        self.emit("song-started", self.song)

        if self.song is not None:
            self.volume = self.volume
            if gapless and self._supports_gapless:
                xine_set_param(self._stream, XINE_PARAM_GAPLESS_SWITCH, 1)
            xine_open(self._stream, self.song("~uri").encode("ascii"))
            if self._paused:
                self._pause()
            else:
                if song is None:
                    self.emit("unpaused")
                self._play()
            if gapless and self._supports_gapless:
                xine_set_param(self._stream, XINE_PARAM_GAPLESS_SWITCH, 0)
        else:
            self.paused = True
            xine_stop(self._stream)

        # seekable might change if we change to None, so notify just in case
        self.notify("seekable")

    def setup(self, playlist, song, seek_pos, explicit=True):
        super().setup(playlist, song, seek_pos, explicit=explicit)
        # xine's declining to seek so soon after startup; try again in 100ms
        if seek_pos:
            GLib.timeout_add(100, self.seek, seek_pos)

    @property
    def eq_bands(self):
        # These are taken straight from Xine's API
        return [30, 60, 125, 250, 500, 1000, 2000, 4000, 8000, 16000]

    def update_eq_values(self):
        bands = self.eq_bands
        need_eq = any(self._eq_values)
        for band, val in enumerate(self._eq_values):
            param = getattr(cdefs, "XINE_PARAM_EQ_%dHZ" % bands[band])
            # between 1..200; 100 is the default gain; 0 means no EQ filter
            # only negative gain seems to work
            val = (int(val * 100 / 24.0) + 100) or 1
            val *= int(need_eq)
            xine_set_param(self._stream, param, val)

    def can_play_uri(self, uri):
        for plugin in self._handle.list_input_plugins():
            if uri.startswith(plugin.lower()):
                return True
        return False


def init(librarian):
    """May raise PlayerError"""

    try:
        driver = config.getbytes("settings", "xine_driver")
    except Exception:
        driver = None
    return XinePlaylistPlayer(driver, librarian)



# Copyright 2019-2020 Joschua Gandert
#                2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
from enum import IntEnum

from gi.repository import Gtk

from quodlibet import _, app, config, qltk, const
from quodlibet.qltk import Icons, ErrorMessage, Message
from quodlibet.plugins import PluginConfig, BoolConfProp, IntConfProp
from quodlibet.plugins.events import EventPlugin
from quodlibet.plugins.songshelpers import is_writable
from quodlibet.util import print_e
from quodlibet.util.songwrapper import SongWrapper


class UpdateStrategy(IntEnum):
    AFTER_PLAY_NOT_SKIP = 0
    AFTER_PLAY_OR_SKIP = 1
    ONCE_ALBUM_RATED = 2


class StrategyText:
    def __init__(self, name, description):
        self.name = name
        self.description = description


STRATEGY_TEXTS = {  #
    UpdateStrategy.AFTER_PLAY_NOT_SKIP: StrategyText(
        _("After every play (default)"),
        _(
            "Whenever a song was played but not skipped, "
            "the plugin will write the tags to the file. "
            "Skip counts aren't stored in files at all, "
            "so this avoids unnecessary writes."
        ),
    ),
    UpdateStrategy.AFTER_PLAY_OR_SKIP: StrategyText(
        _("After every play or skip"),
        _(
            "Whenever a song was played or skipped, "
            "the plugin will write the tags to the file. "
            "Can be useful if you want to make sure that ratings of songs "
            "you dislike and thus skipped are written to the files."
        ),
    ),
    UpdateStrategy.ONCE_ALBUM_RATED: StrategyText(
        _("Once, when album fully rated"),
        _(
            "When a song was played or skipped, the album of that song will "
            "be checked. If every song in the album has been rated and at "
            "least one has no ratings or play counts stored in its file, "
            "the plugin will write the tags to the songs' files.\n\n"
            "Use this to avoid constant file updates, "
            "but be aware that once an album was updated, "
            "you'll have to use the 'Update Tags in Files' plugin "
            "whenever you want modified ratings and play counts "
            "to be written to the files."
        ),
    ),
}

PLAY_COUNT_ABOVE_ZERO_TOOLTIP = _(
    "When the plugin writes the tags of an album, "
    "it will first set the play count of the songs which are zero to one.\n"
    "Sometimes you already know that you don't like a song, "
    "so setting it to one when saving can be useful later on, "
    "when searching for albums you have fully listened to (%s)."
)

WRITE_ERROR_FMT = _("Couldn't write '%s'")


class Config:
    _config = PluginConfig("autoupdatetagsinfiles")

    update_strategy = IntConfProp(
        _config, "update_strategy", UpdateStrategy.AFTER_PLAY_NOT_SKIP.value
    )
    ensure_play_counts_above_zero = BoolConfProp(  # useful for searching
        _config, "ensure_play_counts_above_zero", False
    )


CONFIG = Config()


class AutoUpdateTagsInFiles(EventPlugin):
    PLUGIN_ID = "AutoUpdateTagsInFiles"
    PLUGIN_NAME = _("Auto Update Tags in Files")
    PLUGIN_DESC = _(
        "When songs were played, update the tags in their files. "
        "This will ensure play counts and ratings are up to date."
    )
    PLUGIN_ICON = Icons.DOCUMENT_SAVE

    def PluginPreferences(self, _):
        return AutoUpdateTagsPrefs()

    def enabled(self):
        if not config.getboolean("editing", "save_to_songs"):
            config.set("editing", "save_to_songs", True)

            warning_text = _(
                "The following setting was enabled as it's "
                "required for this plugin to work:\n\n%s"
            )
            setting_name = _("Save ratings and play _counts in tags")

            Message(
                Gtk.MessageType.INFO,
                app.window,
                _("Settings updated"),
                warning_text % setting_name.replace("_", ""),
            ).run()

    def plugin_on_song_ended(self, song, skipped):
        if song is None or not is_writable(song):
            return

        strategy = CONFIG.update_strategy
        if strategy == UpdateStrategy.AFTER_PLAY_NOT_SKIP:
            if not skipped:
                self._try_to_update_song(song)
            return
        if strategy == UpdateStrategy.AFTER_PLAY_OR_SKIP:
            self._try_to_update_song(song)
            return

        self._update_album_if_fully_rated(song.album_key)

    def _try_to_update_song(self, song_wrapper):
        try:
            song_wrapper._needs_write = True
        except Exception as e:
            print_e(e)
            self._error_msg(WRITE_ERROR_FMT % song_wrapper._song)

    def _error_msg(self, message):
        title = _("Error in %s") % self.PLUGIN_NAME
        ErrorMessage(app.window, title, message).run()

    def _update_album_if_fully_rated(self, album_key):
        album = app.library.albums.get(album_key, None)
        if album is None:
            return

        songs = album.songs
        # first check for ratings to avoid costly file checks
        if not songs or not all(song.has_rating for song in songs):
            return

        songs = [s for s in songs if s.supports_rating_and_play_count_in_file]
        if not songs:
            return

        email = config.get("editing", "save_email", const.EMAIL).strip()
        if all(s.has_rating_and_playcount_in_file(email) for s in songs):
            return

        # at least one song has no ratings or play counts stored in files
        try:
            self._update_album(songs)
        except Exception as e:
            print_e(e)
            self._error_msg(WRITE_ERROR_FMT % album_key)

    def _update_album(self, songs):
        song_wrappers = []

        req_play_counts_above_zero = CONFIG.ensure_play_counts_above_zero
        for song in songs:
            if req_play_counts_above_zero and not song.get("~#playcount", 0):
                song["~#playcount"] = 1

            wrapper = SongWrapper(song)
            wrapper._needs_write = True
            song_wrappers.append(wrapper)


class AutoUpdateTagsPrefs(Gtk.Box):
    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        strategy_boxes = []
        for strategy in UpdateStrategy:
            desc = STRATEGY_TEXTS[strategy].description
            desc_label = Gtk.Label(label=desc, wrap=True)

            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
            box.pack_start(desc_label, False, False, 0)

            strategy_boxes.append(box)

        def ensure_play_count_toggled(button, *args):
            CONFIG.ensure_play_counts_above_zero = button.get_active()

        ensure_play_count_checkbutton = Gtk.CheckButton(
            label=_("Ensure play counts are above zero when saving"),
            valign=Gtk.Align.START,
        )
        ensure_play_count_checkbutton.set_tooltip_text(
            PLAY_COUNT_ABOVE_ZERO_TOOLTIP % "#(playcount:min = 1)"
        )
        ensure_play_count_checkbutton.set_active(CONFIG.ensure_play_counts_above_zero)
        ensure_play_count_checkbutton.connect("toggled", ensure_play_count_toggled)

        album_box = strategy_boxes[UpdateStrategy.ONCE_ALBUM_RATED]
        album_box.pack_start(ensure_play_count_checkbutton, False, False, 0)

        def show_only_current_box():
            current = CONFIG.update_strategy
            for n, box in enumerate(strategy_boxes):
                box.set_visible(n == current)

        grid = Gtk.Grid(column_spacing=6, row_spacing=6)

        def grid_add(x, y, child):
            grid.attach(child, x, y, 1, 1)

        def change_strategy(button):
            new_strategy = button.get_active()
            CONFIG.update_strategy = new_strategy
            show_only_current_box()

        update_combobox = Gtk.ComboBoxText()
        for strategy in UpdateStrategy:
            update_combobox.append_text(STRATEGY_TEXTS[strategy].name)

        update_combobox.set_active(CONFIG.update_strategy)
        update_combobox.connect("changed", change_strategy)

        update_lbl = ConfigLabel(_("_Update strategy:"), update_combobox)

        grid_add(0, 0, update_lbl)
        grid_add(1, 0, update_combobox)
        for box in strategy_boxes:
            # show_all will be called on the plugin preference interface, so
            # without the following that would result in overlapping text
            box.show_all()
            box.props.no_show_all = True

            grid_add(1, 1, box)

        show_only_current_box()

        frame = qltk.Frame(label=_("Preferences"), child=grid)
        self.pack_start(frame, False, False, 0)


class ConfigLabel(Gtk.Label):
    """Customised Label for configuration, tied to a widget"""

    def __init__(self, text, widget):
        super().__init__(label=text, use_underline=True)
        self.set_mnemonic_widget(widget)
        self.set_alignment(0.0, 0.5)



# Copyright 2004-2009 Joe Wreschnig, Michael Urman, Steven Robertson
#           2011-2024 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import re
import stat
import sys
import errno
import codecs
import shlex
from typing import Any
from urllib.parse import urlparse, quote, unquote

from gi.repository import GLib

from senf import fsnative, bytes2fsn, fsn2bytes, fsn2text, path2fsn, uri2fsn, _fsnative

from . import windows
from .environment import is_windows
from .misc import NamedTemporaryFile

if sys.platform == "darwin":
    from Foundation import NSString


def mkdir(dir_, *args):
    """Make a directory, including all its parent directories. This does not
    raise an exception if the directory already exists (and is a
    directory)."""

    try:
        os.makedirs(dir_, *args)
    except OSError as e:
        if e.errno != errno.EEXIST or not os.path.isdir(dir_):
            raise


def uri2gsturi(uri):
    """Takes a correct uri and returns a gstreamer-compatible uri"""
    if not is_windows():
        return uri
    try:
        # gstreamer requires extra slashes for network shares
        return GLib.filename_to_uri(uri2fsn(uri))
    except (GLib.Error, ValueError):
        return uri


def iscommand(s):
    """True if an executable file `s` exists in the user's path, or is a
    fully qualified and existing executable file."""

    if s == "" or os.path.sep in s:
        return os.path.isfile(s) and os.access(s, os.X_OK)
    s = s.split()[0]
    path = os.environ.get("PATH", "") or os.defpath
    for p in path.split(os.path.pathsep):
        p2 = os.path.join(p, s)
        if os.path.isfile(p2) and os.access(p2, os.X_OK):
            return True
    else:
        return False


def listdir(path, hidden=False):
    """List files in a directory, sorted, fully-qualified.

    If hidden is false, Unix-style hidden files are not returned.
    """

    assert isinstance(path, fsnative)

    if hidden:
        filt = None
    else:

        def filt(base):
            return not base.startswith(".")

    if path.endswith(os.sep):
        join = "".join
    else:
        join = os.sep.join
    return [
        join([path, basename])
        for basename in sorted(os.listdir(path))
        if filt(basename)
    ]


def mtime(filename):
    """Return the mtime of a file, or 0 if an error occurs."""
    try:
        return os.path.getmtime(filename)
    except OSError:
        return 0


def filesize(filename):
    """Return the size of a file, or 0 if an error occurs."""
    try:
        return os.path.getsize(filename)
    except OSError:
        return 0


def escape_filename(s: str, safe: bytes = b""):
    """Escape a string in a manner suitable for a filename.

    Args:
        s (str) The string to convert
        safe (bytes) A string of characters that needn't be quoted
    Returns:
        fsnative
    """

    s = str(s)
    quoted = quote(s, safe=safe, encoding="utf-8")
    if isinstance(quoted, bytes):
        return bytes2fsn(quoted, "utf-8")
    return bytes2fsn(quoted.encode("ascii"), "utf-8")


def unescape_filename(filename: _fsnative) -> str:
    """Unescape a string in a manner suitable for a filename.

    Args:
        filename (fsnative)
    Returns:
        str
    """

    assert isinstance(filename, fsnative)
    return fsn2text(unquote(filename))


def join_path_with_escaped_name_of_legal_length(path: str, stem: str, ext: str) -> str:
    """Returns a path joined with the escaped stem and the unescaped extension.
    Stem is trimmed until the filename fits into the filesystems maximum file length"""

    # returns the maximum possible filename length at path (subtract one for dot)
    max_stem_length = os.pathconf(path, "PC_NAME_MAX") - 1 - len(ext)

    escaped_stem = escape_filename(stem)
    while len(escaped_stem) > max_stem_length:
        # We don't want to break the escaping, so we only trim the actual name
        stem = stem[:max_stem_length]
        max_stem_length -= 1
        escaped_stem = escape_filename(stem)

    return os.path.join(path, f"{escaped_stem}.{ext}")


def stem_of_file_name(file_name: str) -> str:
    """:return: file name without the extension.

    Note these examples showcasing edge cases:

    >>> stem_of_file_name('a.b.c')
    'a.b'
    >>> stem_of_file_name('.test')
    '.test'
    """
    return os.path.splitext(file_name)[0]


def extension_of_file_name(file_name: str) -> str:
    """:return: extension of the file name. Is empty, or starts with a period.

    Note these examples showcasing edge cases:

    >>> extension_of_file_name('a.b.c')
    '.c'
    >>> extension_of_file_name('.test')
    ''
    """
    return os.path.splitext(file_name)[-1]


def unexpand(filename):
    """Replace the user's home directory with ~ or %USERPROFILE%, if it
    appears at the start of the path name.

    Args:
        filename (fsnative): The file path
    Returns:
        fsnative: The path with the home directory replaced
    """

    sub = (os.name == "nt" and fsnative("%USERPROFILE%")) or fsnative("~")
    home = os.path.normcase(get_home_dir()).rstrip(os.path.sep)
    norm = os.path.normcase(filename)
    if norm == home:
        return sub
    if norm.startswith(home + os.path.sep):
        filename = sub + filename[len(home) :]
    return filename


if is_windows():

    def ismount(path):
        # this can raise on py3+win, but we don't care
        try:
            return os.path.ismount(path)
        except OSError:
            return False
else:
    ismount = os.path.ismount


def find_mount_point(path):
    while not ismount(path):
        path = os.path.dirname(path)
    return path


def xdg_get_system_data_dirs():
    """http://standards.freedesktop.org/basedir-spec/latest/"""

    if os.name == "nt":
        from gi.repository import GLib

        dirs = []
        for dir_ in GLib.get_system_data_dirs():
            dirs.append(dir_)
        return dirs

    data_dirs = os.getenv("XDG_DATA_DIRS")
    if data_dirs:
        return [os.path.abspath(d) for d in data_dirs.split(":")]
    return ("/usr/local/share/", "/usr/share/")


def xdg_get_cache_home():
    if os.name == "nt":
        from gi.repository import GLib

        return GLib.get_user_cache_dir()

    data_home = os.getenv("XDG_CACHE_HOME")
    if data_home:
        return os.path.abspath(data_home)
    return os.path.join(os.path.expanduser("~"), ".cache")


def xdg_get_data_home():
    if os.name == "nt":
        from gi.repository import GLib

        return GLib.get_user_data_dir()

    data_home = os.getenv("XDG_DATA_HOME")
    if data_home:
        return os.path.abspath(data_home)
    return os.path.join(os.path.expanduser("~"), ".local", "share")


def xdg_get_config_home():
    if os.name == "nt":
        from gi.repository import GLib

        return GLib.get_user_config_dir()

    data_home = os.getenv("XDG_CONFIG_HOME")
    if data_home:
        return os.path.abspath(data_home)
    return os.path.join(os.path.expanduser("~"), ".config")


def parse_xdg_user_dirs(data):
    """Parses xdg-user-dirs and returns a dict of keys and paths.

    The paths depend on the content of environ while calling this function.
    See http://www.freedesktop.org/wiki/Software/xdg-user-dirs/

    Args:
        data (bytes)

    Can't fail (but might return garbage).
    """

    assert isinstance(data, bytes)

    paths = {}
    for line in data.splitlines():
        if line.startswith(b"#"):
            continue
        parts = line.split(b"=", 1)
        if len(parts) <= 1:
            continue
        key = parts[0]
        try:
            values = shlex.split(bytes2fsn(parts[1], "utf-8"))
        except ValueError:
            continue
        if len(values) != 1:
            continue
        paths[key] = os.path.normpath(os.path.expandvars(values[0]))

    return paths


def xdg_get_user_dirs():
    """Returns a dict of xdg keys to paths. The paths don't have to exist."""
    config_home = xdg_get_config_home()
    try:
        with open(os.path.join(config_home, "user-dirs.dirs"), "rb") as h:
            return parse_xdg_user_dirs(h.read())
    except OSError:
        return {}


def get_temp_cover_file(data: bytes, mime: str | None = None) -> Any:
    """Returns a file object or None"""

    try:
        suffix = None
        if mime:
            mime = mime.lower()
            if "png" in mime:
                suffix = fsnative(".png")
            elif "jpg" in mime or "jpeg" in mime:
                suffix = fsnative(".jpg")
        # pass fsnative so that mkstemp() uses unicode on Windows
        fn = NamedTemporaryFile(prefix=fsnative("cover-"), suffix=suffix)
        fn.write(data)
        fn.flush()
        fn.seek(0, 0)
    except OSError:
        return None
    else:
        return fn


def _strip_win32_incompat(string, bad=r'\:*?;"<>|'):
    """Strip Win32-incompatible characters from a Windows or Unix path."""

    if os.name == "nt":
        bad += "/"

    if not string:
        return string

    new = "".join((s in bad and "_") or s for s in string)
    parts = new.split(os.sep)

    def fix_end(string):
        return re.sub(r"[\. ]$", "_", string)

    return os.sep.join(fix_end(p) for p in parts)


def strip_win32_incompat_from_path(string):
    """Strip Win32-incompatible chars from a path, ignoring os.sep
    and the drive part"""

    drive, tail = os.path.splitdrive(string)
    tail = os.sep.join(_strip_win32_incompat(s) for s in tail.split(os.sep))
    return drive + tail


def _normalize_darwin_path(filename, canonicalise=False):
    filename = path2fsn(filename)

    if canonicalise:
        filename = os.path.realpath(filename)
    filename = os.path.normpath(filename)

    data = fsn2bytes(filename, "utf-8")
    decoded = data.decode("utf-8", "quodlibet-osx-path-decode")

    try:
        return bytes2fsn(NSString.fileSystemRepresentation(decoded), "utf-8")
    except ValueError:
        return filename


def _normalize_path(filename, canonicalise=False):
    """Normalize a path on Windows / Linux
    If `canonicalise` is True, dereference symlinks etc
    by calling `os.path.realpath`
    """
    filename = path2fsn(filename)
    if canonicalise:
        filename = os.path.realpath(filename)
    filename = os.path.normpath(filename)
    return os.path.normcase(filename)


if sys.platform == "darwin":

    def _osx_path_decode_error_handler(error):
        bytes_ = bytearray(error.object[error.start : error.end])
        return "".join("%%%X".__mod__(b) for b in bytes_), error.end

    codecs.register_error("quodlibet-osx-path-decode", _osx_path_decode_error_handler)

    normalize_path = _normalize_darwin_path
else:
    normalize_path = _normalize_path


def path_equal(p1, p2, canonicalise=False):
    return normalize_path(p1, canonicalise) == normalize_path(p2, canonicalise)


def limit_path(path, ellipsis=True):
    """Reduces the filename length of all filenames in the given path
    to the common maximum length for current platform.

    While the limits are depended on the file system and more restrictions
    may apply, this covers the common case.
    """

    assert isinstance(path, fsnative)

    main, ext = os.path.splitext(path)
    parts = main.split(os.sep)
    for i, p in enumerate(parts):
        # Limit each path section to 255 (bytes on linux, chars on win).
        # http://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits
        limit = 255
        if i == len(parts) - 1:
            limit -= len(ext)

        if len(p) > limit:
            if ellipsis:
                p = p[: limit - 2] + fsnative("..")
            else:
                p = p[:limit]
        parts[i] = p

    return os.sep.join(parts) + ext


def get_home_dir():
    """Returns the root directory of the user, /home/user or C:\\Users\\user"""

    if os.name == "nt":
        return windows.get_profile_dir()
    return os.path.expanduser("~")


def is_hidden(path: _fsnative) -> bool:
    """Returns if a directory / file is considered hidden by the platform.

    Hidden meaning the user should normally not be exposed to those files when
    opening the parent directory in the default file manager using the default
    settings.

    Does not check if any of the parents are hidden.
    If the file / dir does not exist, the result is implementation defined.

    :param path: the path to check
    :return: True if and only if the path is considered hidden on the system
    """

    if sys.platform == "windows":
        return bool(os.stat(path).st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN)
    basename = os.path.basename(path)
    # Let's allow "...and Justice For All" etc (#3916)
    return basename.startswith(".") and basename[1:2] != "."


def uri_is_valid(uri):
    """Returns True if the passed in text is a valid URI (file, http, etc.)

    Args:
        uri(text or bytes)
    Returns:
        bool
    """

    try:
        if isinstance(uri, bytes):
            uri.decode("ascii")
        elif not isinstance(uri, bytes):
            uri = uri.encode("ascii")
    except ValueError:
        return False

    parsed = urlparse(uri)
    if not parsed.scheme or not len(parsed.scheme) > 1:
        return False
    if not (parsed.netloc or parsed.path):
        return False
    return True


class RootPathFile:
    """Simple container used for discerning a pathfile's 'root' directory
    and 'end' part. The variable depth of a pathfile's 'end' part renders
    os.path built-ins (basename etc.) useless for this purpose"""

    _root = ""  # 'root' of full file path
    _pathfile = ""  # full file path

    def __init__(self, root, pathfile):
        self._root = root
        self._pathfile = pathfile

    @property
    def root(self):
        return self._root

    @property
    def end(self):
        return self._pathfile[len(self._root) + len(os.sep) :]

    @property
    def pathfile(self):
        return self._pathfile

    @property
    def end_escaped(self):
        escaped = [escape_filename(part) for part in self.end.split(os.path.sep)]
        return os.path.sep.join(escaped)

    @property
    def pathfile_escaped(self):
        return os.path.sep.join([self.root, self.end_escaped])

    @property
    def valid(self):
        valid = True
        if os.path.exists(self.pathfile):
            return valid
        try:
            with open(self.pathfile, "w", encoding="utf-8") as f:
                f.close()  # do nothing
        except OSError:
            valid = False
        if os.path.exists(self.pathfile):
            os.remove(self.pathfile)
        return valid



# Copyright 2007-2008 Joe Wreschnig
#           2014 Christoph Reiter
#      2014-2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import _, ngettext

"""Database of all known tags, their translations and how they are used"""


class TagName:
    """
    Text:
        desc -- translated description
        plural -- translated plural description or None
        role -- translated role description (for people tags)

    Types:
        user -- user editable tag e.g. "foo"
            hidden -- if user tag should only be shown in the tag editor (e.g.
                      tracknumber is hidden, since ~#track/~#tracks are more
                      useful for displaying)
            has_sort -- has a sort user variant e.g. "foosort"
            machine -- user tag is not human readable. Something people might
                       want to hide in the tag editor.

        internal -- generated by QL e.g. "~foo"
            hidden -- if it is replaced by another tag (for backwards compat)
            has_sort -- has a sort user variant e.g. "~foosort"
            has_roles -- has a roles variant e.g. "~foo:roles"

        numeric -- e.g. "~#foo"
    """

    def __init__(self, name, options, desc, plural=None, role=None):
        # I don't think this categorization is any good.. but at least
        # there is any..
        self.name = name
        self.desc = desc
        self.plural = plural
        self.user = "u" in options
        self.internal = "i" in options
        self.numeric = "n" in options
        self.machine = "m" in options
        self.has_sort = "s" in options
        self.has_roles = "r" in options
        self.hidden = "h" in options
        self.role = role

        # some sanity checks
        assert self.user or self.internal or self.numeric
        assert not (set(options) - set("uinmshr"))

        if self.has_roles:
            assert self.internal
        if self.has_sort:
            assert self.user or self.internal
        if self.machine:
            assert self.user
        if self.hidden:
            assert self.user + self.internal == 1

    def __repr__(self):
        return f"{type(self).__name__}({vars(self)!r})"


def _get_role_map(tags):
    roles = {}
    for name, tag in tags.items():
        if tag.role:
            roles[name] = tag.role
            if tag.has_sort:
                roles[name + "sort"] = tag.role
    return roles


T = TagName
_TAGS = {
    t.name: t
    for t in [
        T("album", "us", _("album"), _("albums")),
        T("arranger", "u", _("arranger"), _("arrangers"), _("arrangement")),
        T("artist", "us", _("artist"), _("artists")),
        T("author", "u", _("author"), _("authors")),
        T("comment", "u", _("comment")),
        T("composer", "us", _("composer"), _("composers"), _("composition")),
        # Translators: conducting as in conducting a musical performance
        T("conductor", "u", _("conductor"), _("conductors"), _("conducting")),
        T("contact", "u", _("contact")),
        T("copyright", "u", _("copyright")),
        T("date", "u", _("date")),
        T("description", "u", _("description")),
        T("genre", "u", _("genre"), _("genres")),
        T("grouping", "u", _("grouping")),
        T("language", "ui", _("language")),
        T("license", "u", _("license")),
        T("location", "u", _("location")),
        T("lyricist", "u", _("lyricist"), _("lyricists"), _("lyrics")),
        # Translators: Also e.g. "record label", "publisher"
        T("organization", "u", _("organization")),
        T("performer", "uisr", _("performer"), _("performers"), _("performance")),
        T("producer", "u", _("producer"), _("producers"), _("production")),
        T("title", "u", _("title")),
        T("version", "u", _("version")),
        T("website", "u", _("website")),
        T("albumartist", "us", _("album artist")),
        T("bpm", "u", _("BPM")),
        T("isrc", "u", "ISRC"),
        # Translators: This used to be called "part".
        T("discsubtitle", "u", _("disc subtitle")),
        T("part", "u", _("disc subtitle")),
        T("discnumber", "uh", _("disc")),
        T("tracknumber", "uh", _("track")),
        T("labelid", "u", _("label ID")),
        T("originaldate", "u", _("original release date")),
        T("originalalbum", "u", _("original album")),
        T("originalartist", "us", _("original artist")),
        T("recordingdate", "u", _("recording date")),
        T("releasecountry", "u", _("release country")),
        T("initialkey", "u", _("initial key")),
        # for backwards compat
        T("performers", "ishr", _("performers")),
        # http://musicbrainz.org/doc/MusicBrainzTag
        # Note: picard has changed musicbrainz_trackid to mean release track.
        # We can't do that because of existing libraries, so use a new
        # musicbrainz_releastrackid instead.
        T("musicbrainz_trackid", "um", _("MusicBrainz recording ID")),
        T("musicbrainz_releasetrackid", "um", _("MusicBrainz release track ID")),
        T("musicbrainz_albumid", "um", _("MusicBrainz release ID")),
        T("musicbrainz_artistid", "um", _("MusicBrainz artist ID")),
        T("musicbrainz_albumartistid", "um", _("MusicBrainz release artist ID")),
        T("musicbrainz_trmid", "um", _("MusicBrainz TRM ID")),
        T("musicip_puid", "um", _("MusicIP PUID")),
        T("musicbrainz_albumstatus", "um", _("MusicBrainz album status")),
        T("musicbrainz_albumtype", "um", _("MusicBrainz album type")),
        T("musicbrainz_releasegroupid", "um", _("MusicBrainz release group ID")),
        # Translators: "gain" means a volume adjustment, not "to acquire".
        T("replaygain_track_gain", "umn", _("track gain")),
        T("replaygain_track_peak", "umn", _("track peak")),
        # Translators: "gain" means a volume adjustment, not "to acquire".
        T("replaygain_album_gain", "umn", _("album gain")),
        T("replaygain_album_peak", "umn", _("album peak")),
        T("replaygain_reference_loudness", "umn", _("reference loudness")),
        T("added", "n", _("added")),
        T("lastplayed", "n", _("last played")),
        T("disc", "n", _("disc")),
        T("discs", "n", _("discs")),
        T("track", "n", _("track")),
        T("tracks", "n", _("tracks")),
        T("laststarted", "n", _("last started")),
        T("filename", "i", _("full name")),
        T("basename", "i", _("filename")),
        T("dirname", "i", _("directory")),
        T("mtime", "n", _("modified")),
        T("playcount", "n", _("plays")),
        T("skipcount", "n", _("skips")),
        T("uri", "i", "URI"),
        T("mountpoint", "i", _("mount point")),
        T("length", "n", _("length")),
        T("people", "isr", _("people")),
        T("rating", "in", _("rating")),
        T("year", "in", _("year")),
        T("originalyear", "in", _("original release year")),
        T("bookmark", "i", _("bookmark")),
        T("bitdepth", "n", _("bitdepth")),
        T("bitrate", "in", _("bitrate")),
        T("filesize", "n", _("file size")),
        T("format", "i", _("file format")),
        T("codec", "i", _("codec")),
        T("encoding", "i", _("encoding")),
        T("playlists", "i", _("playlists")),
        T("samplerate", "n", _("sample rate")),
        T("channels", "n", _("channel count")),
    ]
}


def _get_sort_map(tags):
    """See TAG_TO_SORT"""

    tts = {}
    for name, tag in tags.items():
        if tag.has_sort:
            if tag.user:
                tts[name] = f"{name}sort"
            if tag.internal:
                tts[f"~{name}"] = f"~{name}sort"
    return tts


def _get_standard_tags(tags, machine=False):
    stags = []
    for name, tag in tags.items():
        if tag.user and tag.machine == machine:
            stags.append(name)
            if tag.has_sort:
                stags.append(f"{name}sort")
    return stags


TAG_TO_SORT = _get_sort_map(_TAGS)
"""A mapping of tag -> sorttag. e.g. artist -> artistsort"""

MACHINE_TAGS = _get_standard_tags(_TAGS, machine=True)
"""A sequence of editable tags that are not human-readable.
e.g. musicbrainz_albumid
"""

USER_TAGS = _get_standard_tags(_TAGS, machine=False)
"""A sequence of tags that are human-readable and can be edited.
e.g. album
"""

TAG_ROLES = _get_role_map(_TAGS)
"""A mapping from tags to their translated role description.
e.g. conductor -> conducting
"""


def readable(tag, plural=False):
    """Gives a translated description for a tag.

    Also supports internal, numeric tags.
    If plural is True, will return a plural description if possible.

    album -> album
    albumsort -> album (sort)
    ~foo -> foo
    ~people:roles -> people (roles)
    """

    try:
        if tag[0] == "~":
            if tag[1] == "#":
                tag = tag[2:]
            else:
                tag = tag[1:]
    except IndexError:
        return ngettext("Invalid tag", "Invalid tags", 1)

    def desc(tag):
        if plural:
            plural_desc = _TAGS[tag].plural
            if plural_desc:
                return plural_desc
        return _TAGS[tag].desc

    if tag in _TAGS:
        return desc(tag)
    if tag == "people:real":
        return desc("people")
    roles = False
    if tag.endswith(":roles"):
        roles = True
        tag = tag[:-6]

    parts = []
    if tag.endswith("sort"):
        v = _TAGS.get(tag[:-4])
        if v is not None and v.has_sort:
            tag = tag[:-4]
            # Translators: e.g. "artist (sort)"
            parts.append(_("sort"))
    else:
        v = _TAGS.get(tag[:-4])

    if roles:
        v = _TAGS.get(tag)
        if v is not None and v.has_roles:
            # Translators: e.g. "performer (roles)"
            parts.append(_("roles"))

    if tag in _TAGS:
        desc = desc(tag)
        if parts:
            desc += " ({})".format(", ".join(parts))
        return desc

    return tag


def sortkey(tag):
    """Sort key for sorting tag names by importance.

    tags.sort(key=sortkey)
    """

    # last one -> most important
    order = [
        "album",
        "artist",
        "title",
    ]

    try:
        return (-order.index(tag), tag)
    except ValueError:
        if tag in MACHINE_TAGS:
            return (2, tag)
        return (1, tag)



# Copyright 2006 Joe Wreschnig
#           2021 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import wave

from senf import fsn2text

from ._audio import AudioFile, translate_errors


extensions = [".wav"]


class WAVEFile(AudioFile):
    format = "WAVE"
    mimes = ["audio/wav", "audio/x-wav", "audio/wave"]

    def __init__(self, filename):
        with translate_errors():
            with open(filename, "rb") as h:
                f = wave.open(h)
                self["~#length"] = float(f.getnframes()) / f.getframerate()
                self["~#channels"] = f.getnchannels()
                self["~#samplerate"] = f.getframerate()
                self["~#bitdepth"] = f.getsampwidth() * 8
        self.sanitize(filename)

    def sanitize(self, filename=None):
        super().sanitize(filename)
        self["title"] = fsn2text(
            os.path.splitext(os.path.basename(self["~filename"]))[0]
        )

    def write(self):
        pass

    def can_change(self, k=None):
        if k is None:
            return ["artist"]
        return k == "artist"


loader = WAVEFile
types = [WAVEFile]



# Copyright 2015    Christoph Reiter
#           2016-23 Nick Boultbee
#           2019    Peter Strulo
#           2022-23 Jej@github
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from collections.abc import Callable

from gi.repository import Gtk

from quodlibet import _
from quodlibet import config
from quodlibet.qltk import Icons
from quodlibet.qltk.ccb import ConfigSwitch
from quodlibet.qltk.entry import UndoEntry
from quodlibet.util.string import decode


def _config(section, option, label, tooltip=None, getter=None):
    def on_changed(entry, *args):
        config.settext(section, option, entry.get_text())

    entry = UndoEntry()
    if tooltip:
        entry.set_tooltip_text(tooltip)
    entry.set_text(config.gettext(section, option))
    entry.connect("changed", on_changed)

    def on_reverted(*args):
        config.reset(section, option)
        entry.set_text(config.gettext(section, option))

    revert = revert_button(on_reverted)
    lbl = Gtk.Label(label=label, use_underline=True)
    lbl.set_mnemonic_widget(entry)
    return lbl, entry, revert


def text_config(section, option, label, tooltip=None):
    def getter(section, option):
        return decode(config.get(section, option))

    return _config(section, option, label, tooltip, getter)


def boolean_config(section, option, label, tooltip):
    def on_reverted(*args):
        config.reset(section, option)
        button.set_active(config.getboolean(section, option))

    button = ConfigSwitch(None, section, option, tooltip=tooltip, populate=True)
    button_box = Gtk.VBox(homogeneous=True)
    button_box.pack_start(button, False, False, 0)

    lbl = Gtk.Label(label=label, use_underline=True)
    lbl.set_mnemonic_widget(button.switch)
    revert = revert_button(on_reverted)
    return lbl, button_box, revert


def revert_button(on_reverted: Callable[..., None]) -> Gtk.Button:
    revert = Gtk.Button()
    revert.add(Gtk.Image.new_from_icon_name(Icons.DOCUMENT_REVERT, Gtk.IconSize.BUTTON))
    revert.connect("clicked", on_reverted)
    revert.set_tooltip_text(_("Revert to default"))
    return revert


def int_config(section, option, label, tooltip):
    def getter(section, option):
        return str(config.getint(section, option))

    return _config(section, option, label, tooltip, getter)


def slider_config(
    section,
    option,
    label,
    tooltip,
    lower=0,
    upper=1,
    on_change_callback=None,
    label_value_callback=None,
):
    def on_reverted(*args):
        config.reset(section, option)
        scale.set_value(config.getfloat(section, option))

    def on_change(scale):
        value = scale.get_value()
        if on_change_callback:
            value = on_change_callback(value)
        scale.set_value(value)
        config.set(section, option, value)

    default = config.getfloat(section, option)

    adjustment = Gtk.Adjustment(value=default, lower=lower, upper=upper)
    scale = Gtk.HScale.new(adjustment)
    scale.set_value_pos(Gtk.PositionType.LEFT)
    scale.set_show_fill_level(True)
    scale.set_tooltip_text(_(tooltip))

    if label_value_callback:
        scale.connect("format-value", lambda _, value: label_value_callback(value))
    scale.connect("value-changed", on_change)

    revert = revert_button(on_reverted)
    lbl = Gtk.Label(label=label, use_underline=True)
    lbl.set_mnemonic_widget(scale)
    return lbl, scale, revert


class AdvancedPreferencesPane(Gtk.VBox):
    MARGIN = 12

    def __init__(self, *args):
        super().__init__()

        # We don't use translations as these things are internal
        # and don't want to burden the translators...
        # TODO: rethink translation here? (#3494)
        rows = [
            text_config(
                "editing",
                "id3encoding",
                "ID3 encodings",
                (
                    "ID3 encodings separated by spaces. "
                    "UTF-8 is always tried first, and Latin-1 is always tried last."
                ),
            ),
            text_config(
                "settings",
                "search_tags",
                "Extra search tags",
                (
                    "Tags that get searched in addition to "
                    'the ones present in the song list. Separate with ","'
                ),
            ),
            text_config(
                "editing",
                "multi_line_tags",
                "Multi-line tags",
                (
                    "Tags to consider as multi-line (delimited by \\n) "
                    "rather than multi-valued (comma-separated)"
                ),
            ),
            text_config("settings", "rating_symbol_full", "Rating symbol (full)"),
            text_config("settings", "rating_symbol_blank", "Rating symbol (blank)"),
            text_config(
                "player",
                "backend",
                "Backend",
                "Identifier of the playback backend to use",
            ),
            boolean_config(
                "settings",
                "disable_hints",
                "Disable hints",
                "Disable popup windows (treeview hints)",
            ),
            int_config(
                "browsers",
                "cover_size",
                "Album cover size",
                (
                    "Size of the album cover images in the album list browser "
                    "(restart required)"
                ),
            ),
            text_config(
                "albumart",
                "search_filenames",
                "Album art search filenames",
                "Which specific files are (also) tried for album art",
            ),
            boolean_config(
                "settings",
                "disable_mmkeys",
                "Disable multimedia keys",
                "(restart required)",
            ),
            text_config(
                "settings",
                "window_title_pattern",
                "Main window title",
                (
                    "A (tied) tag for the main window title, e.g. ~title~~people "
                    "(restart required)"
                ),
            ),
            text_config(
                "settings",
                "datecolumn_timestamp_format",
                "Timestamp date format",
                "A timestamp format for dates, e.g. %Y%m%d %X (restart required)",
            ),
            boolean_config(
                "settings",
                "scrollbar_always_visible",
                "Scrollbars always visible",
                (
                    "Toggles whether the scrollbars on the bottom and side of "
                    "the window always are visible or get hidden when not in use "
                    "(restart required)"
                ),
            ),
            boolean_config(
                "settings",
                "monospace_query",
                "Use monospace font for search input",
                "Helps readability of code-like queries, but looks less consistent "
                "(restart required)",
            ),
            text_config(
                "settings",
                "query_font_size",
                "Search input font size",
                "Size to apply to the search query entry, "
                "in any Pango CSS units, e.g. '100%', '1rem'. (restart required)",
            ),
            boolean_config(
                "settings",
                "pangocairo_force_fontconfig",
                "Force Use Fontconfig Backend",
                "It's not the default on win/macOS (restart required)",
            ),
            text_config(
                "browsers",
                "ignored_characters",
                "Ignored characters",
                "Characters to ignore in queries",
            ),
            boolean_config(
                "settings",
                "plugins_window_on_top",
                "Plugin window on top",
                "Toggles whether the plugin window appears on top of others",
            ),
            int_config(
                "autosave",
                "queue_interval",
                "Queue autosave interval",
                (
                    "Longest time between play queue auto-saves, or 0 for disabled. "
                    "(restart required)"
                ),
            ),
            int_config(
                "browsers",
                "searchbar_historic_entries",
                "Number of history entries in the search bar",
                "8 by default (restart advised)",
            ),
            int_config(
                "browsers",
                "searchbar_enqueue_limit",
                "Search bar confirmation limit for enqueue",
                (
                    "Maximal size of the song list that can be enqueued from "
                    "the search bar without confirmation."
                ),
            ),
            slider_config(
                "player",
                "playcount_minimum_length_proportion",
                "Minimum length to consider a track as played",
                (
                    "Consider a track played after listening to this proportion of "
                    "its total duration"
                ),
                label_value_callback=lambda value: f"{int(value * 100)}%",
            ),
            text_config(
                "browsers",
                "missing_title_template",
                "Missing title template string",
                (
                    "Template for building title of tracks when title tag is missing. "
                    "Tags are allowed, like <~basename> <~dirname> <~format> <~length> "
                    "<~#bitrate>, etc. See tags documentation for details."
                ),
            ),
            boolean_config(
                "player",
                "gst_use_playbin3",
                "Use GStreamer playbin3",
                "Use GStreamer playbin3 instead of playbin for playback.",
            ),
        ]

        # Tabulate all settings for neatness
        table = Gtk.Table(n_rows=len(rows), n_columns=4)
        table.set_col_spacings(12)
        table.set_row_spacings(8)
        table.set_no_show_all(True)

        for row, (label, widget, button) in enumerate(rows):
            label.set_alignment(0.0, 0.5)
            table.attach(label, 0, 1, row, row + 1, xoptions=Gtk.AttachOptions.FILL)
            if isinstance(widget, Gtk.VBox):
                # This stops checkbox from expanding too big, or shrinking text entries
                blank = Gtk.Label()
                table.attach(
                    blank, 1, 2, row, row + 1, xoptions=Gtk.AttachOptions.EXPAND
                )
                table.attach(
                    widget, 2, 3, row, row + 1, xoptions=Gtk.AttachOptions.SHRINK
                )
            else:
                xoptions = Gtk.AttachOptions.FILL
                table.attach(widget, 1, 3, row, row + 1, xoptions=xoptions)
            table.attach(button, 3, 4, row, row + 1, xoptions=Gtk.AttachOptions.SHRINK)

        def on_click(button):
            button.hide()
            table.set_no_show_all(False)
            table.show_all()

        button = Gtk.Button(label=_("I know what I'm doing!"), use_underline=True)
        button.connect("clicked", on_click)

        self.pack_start(button, False, True, 0)
        self.pack_start(table, True, True, 0)



# Copyright 2004-2005 Joe Wreschnig, Michael Urman
#           2012 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from ._audio import (
    PEOPLE,
    AudioFile,
    DUMMY_SONG,
    decode_value,
    FILESYSTEM_TAGS,
    TIME_TAGS,
)
from ._image import EmbeddedImage, APICType
from ._misc import AudioFileError, init, MusicFile, types, loaders, filter, mimes
from ._serialize import load_audio_files, dump_audio_files, SerializationError

(
    AudioFile,
    AudioFileError,
    EmbeddedImage,
    DUMMY_SONG,
    PEOPLE,
    decode_value,
)
(
    APICType,
    FILESYSTEM_TAGS,
    TIME_TAGS,
    init,
    MusicFile,
    types,
    loaders,
    filter,
)
mimes, load_audio_files, dump_audio_files, SerializationError



# Copyright 2018 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import print_d
from quodlibet.util import is_linux
from ._base import SessionClient, SessionError


def iter_backends():
    if is_linux():
        from .gnome import GnomeSessionClient

        yield GnomeSessionClient
        from .xfce import XfceSessionClient

        yield XfceSessionClient
        from .xsmp import XSMPSessionClient

        yield XSMPSessionClient
    # dummy one last
    yield SessionClient


def init(app):
    """Returns an active SessionClient instance or None"""

    for backend in iter_backends():
        print_d(f"Trying {backend.__name__}")
        client = backend()
        try:
            client.open(app)
        except SessionError as e:
            print_d(str(e))
        else:
            return client
    raise AssertionError()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
from pathlib import Path
from typing import TypeVar
from collections.abc import Iterable

from quodlibet import util, print_d
from quodlibet.formats import MusicFile, AudioFile
from quodlibet.library.album import AlbumLibrary
from quodlibet.library.base import Library, PicklingMixin, K
from quodlibet.library.file import WatchedFileLibraryMixin
from quodlibet.library.playlist import PlaylistLibrary
from quodlibet.query import Query
from quodlibet.util.path import normalize_path
from senf import fsnative

V = TypeVar("V", bound=AudioFile)


class SongLibrary(Library[K, V], PicklingMixin):
    """A library for songs.

    Items in this kind of library must support (roughly) the AudioFile
    interface.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @util.cached_property
    def albums(self):
        return AlbumLibrary(self)

    @util.cached_property
    def playlists(self):
        pl_lib = PlaylistLibrary(self)
        print_d(f"Created playlist library {pl_lib}")
        return pl_lib

    def destroy(self):
        super().destroy()
        if "albums" in self.__dict__:
            self.albums.destroy()
        if "playlists" in self.__dict__:
            self.playlists.destroy()

    def tag_values(self, tag):
        """Return a set of all values for the given tag."""
        return {value for song in self.values() for value in song.list(tag)}

    def rename(self, song, new_name, changed: set | None = None):
        """Rename a song.

        This requires a special method because it can change the
        song's key.

        The 'changed' signal may fire for this library or the changed
        song is added to the passed changed set().

        If the song exists in multiple libraries you cannot use this
        method. Instead, use the librarian.
        """
        if song.key == new_name:
            print_d(f"Nothing changed for {new_name!r}")
            return
        print_d(f"Renaming {song.key!r} to {new_name!r}", self)
        del self._contents[song.key]
        song.rename(new_name)
        self._contents[song.key] = song
        if changed is not None:
            changed.add(song)
        else:
            self.changed({song})

    def query(self, text, sort=None, star=Query.STAR):
        """Query the library and return matching songs."""
        if isinstance(text, bytes):
            text = text.decode("utf-8")

        songs = self.values()
        if text != "":
            search = Query(text, star).search
            songs = [s for s in songs if search(s)]
        return songs


class SongFileLibrary(SongLibrary, WatchedFileLibraryMixin):
    """A library containing song files.
    Pickles contents to disk as `FileLibrary`"""

    def __init__(self, name=None, watch_dirs: Iterable[fsnative] | None = None):
        print_d(f"Initializing {type(self).__name__}: {name!r}")
        super().__init__(name)
        if watch_dirs:
            self.start_watching(watch_dirs)

    def get_filename(self, filename):
        key = normalize_path(filename, True)
        return self._contents.get(key)

    def add_filename(self, filename: str | Path, add: bool = True) -> AudioFile | None:
        """Add a song to the library based on filename.

        If 'add' is true, the song will be added and the 'added' signal
        may be fired.

        Example (add=False):
            load many songs and call Library.add(songs) to add all in one go.

        The song is returned if it is in the library after this call.
        Otherwise, None is returned.
        """

        key = normalize_path(filename, True)
        song = None
        if key not in self._contents:
            song = MusicFile(filename)
            if song and add:
                self.add([song])
        else:
            print_d(f"Already got file {filename!r}")
            song = self._contents[key]

        return song
